OPTION_MAX_RETRIES = "max_fetch_token_retries"
OPTION_RENEW_REQUIRED = "token_renew_required"
OPTION_RENEW_SYNCHRONOUS = "token_renew_synchronous"
OPTION_REPORT_ERROR = "token_renew_report_error"
OPTION_RENEW_REQUEST_TIMEOUT = "token_renew_request_timeout"



from datetime import timedelta

from nebius.aio.abc import ClientChannelInterface
from nebius.aio.token.deferred_channel import DeferredChannel
from nebius.aio.token.exchangeable import Bearer as ExchangeableBearer
from nebius.aio.token.renewable import Bearer as RenewableBearer
from nebius.aio.token.token import Bearer as ParentBearer
from nebius.aio.token.token import NamedBearer, Receiver
from nebius.base.service_account.federated_credentials import (
    FederatedCredentialsBearer as FederatedCredentialsReader,
)
from nebius.base.service_account.federated_credentials import (
    FederatedCredentialsTokenRequester,
    FileFederatedCredentials,
)


class FederatedCredentialsBearer(ParentBearer):
    def __init__(
        self,
        federated_credentials: (
            FederatedCredentialsTokenRequester | FederatedCredentialsReader | str
        ),
        service_account_id: str | None = None,
        channel: ClientChannelInterface | DeferredChannel | None = None,
        max_retries: int = 2,
        lifetime_safe_fraction: float = 0.9,
        initial_retry_timeout: timedelta = timedelta(seconds=1),
        max_retry_timeout: timedelta = timedelta(minutes=1),
        retry_timeout_exponent: float = 1.5,
        refresh_request_timeout: timedelta = timedelta(seconds=5),
    ) -> None:
        if isinstance(federated_credentials, str):
            federated_credentials = FileFederatedCredentials(federated_credentials)
        if isinstance(federated_credentials, FederatedCredentialsReader):
            if not isinstance(service_account_id, str):
                raise TypeError(
                    "Service account ID must be provided as a string when "
                    "federated_credentials is a string."
                )
            federated_credentials = FederatedCredentialsTokenRequester(
                service_account_id=service_account_id,
                credentials=federated_credentials,
            )

        if not isinstance(federated_credentials, FederatedCredentialsTokenRequester):  # type: ignore[unused-ignore]
            raise TypeError(
                "federated_credentials must be FederatedCredentialsTokenRequester, "
                "FederatedCredentialsBearer or string"
                f", got {type(federated_credentials)}"
            )

        self._exchangeable = ExchangeableBearer(
            federated_credentials,
            channel=channel,
            max_retries=max_retries,
        )
        self._source: ParentBearer = RenewableBearer(
            self._exchangeable,
            max_retries=max_retries,
            lifetime_safe_fraction=lifetime_safe_fraction,
            initial_retry_timeout=initial_retry_timeout,
            max_retry_timeout=max_retry_timeout,
            retry_timeout_exponent=retry_timeout_exponent,
            refresh_request_timeout=refresh_request_timeout,
        )

        if isinstance(federated_credentials.credentials, FileFederatedCredentials):
            self._source = NamedBearer(
                self._source,
                f"federated-credentials/{federated_credentials.service_account_id}"
                f"/{federated_credentials.credentials.file_path}",
            )

    def set_channel(self, channel: ClientChannelInterface) -> None:
        self._exchangeable.set_channel(channel)

    @property
    def wrapped(self) -> "ParentBearer|None":
        return self._source

    def receiver(self) -> "Receiver":
        return self._source.receiver()



from collections.abc import Awaitable
from typing import TypeVar

from nebius.aio.abc import ClientChannelInterface
from nebius.aio.authorization.authorization import Provider as AuthorizationProvider

from .base import AddressChannel

T = TypeVar("T")


class Constant(ClientChannelInterface):
    def __init__(
        self,
        method: str,
        source: ClientChannelInterface,
        parent_id: str | None = None,
    ) -> None:
        self._method = method
        self._parent_id = parent_id or source.parent_id()
        self._source = source

    def return_channel(self, chan: AddressChannel | None) -> None:
        return self._source.return_channel(chan)

    def discard_channel(self, chan: AddressChannel | None) -> None:
        return self._source.discard_channel(chan)

    def parent_id(self) -> str | None:
        return self._parent_id

    def get_authorization_provider(self) -> AuthorizationProvider | None:
        return self._source.get_authorization_provider()

    def get_channel_by_method(self, method_name: str) -> AddressChannel:
        return self._source.get_channel_by_method(self._method)

    def run_sync(self, awaitable: Awaitable[T], timeout: float | None = None) -> T:
        return self._source.run_sync(awaitable, timeout)



DOMAIN: str = "api.nebius.cloud:443"
DEFAULT_CONFIG_DIR: str = "~/.nebius"
DEFAULT_CONFIG_FILE: str = "config.yaml"
DEFAULT_CREDENTIALS_FILE: str = "credentials.yaml"
TOKEN_ENV = "NEBIUS_IAM_TOKEN"  # noqa: S105
PROFILE_ENV = "NEBIUS_PROFILE"



from datetime import datetime, timedelta, timezone
from logging import getLogger
from pathlib import Path

from nebius.aio.token.token import Bearer as ParentBearer
from nebius.aio.token.token import Receiver as ParentReceiver
from nebius.aio.token.token import Token
from nebius.base.constants import DEFAULT_CONFIG_DIR, DEFAULT_CREDENTIALS_FILE

from .throttled_token_cache import ThrottledTokenCache

log = getLogger(__name__)


class RenewableFileCacheReceiver(ParentReceiver):
    def __init__(
        self,
        bearer: "RenewableFileCacheBearer",
        cache: ThrottledTokenCache,
    ) -> None:
        super().__init__()
        self._bearer = bearer
        self._cache = cache
        self._receiver: ParentReceiver | None = None
        self._last_saved: Token | None = None
        self._from_cache: bool = True

    async def _fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        if self._from_cache:
            token = await self._cache.get()
        else:
            token = await self._cache.refresh()
            if self._last_saved == token:
                token = None  # to avoid using the same token if error occurs
        if token is not None and not token.is_expired():
            if self._bearer.safety_margin is None or (
                not token.expiration
                or (
                    token.expiration - self._bearer.safety_margin
                    > datetime.now(timezone.utc)
                )
            ):
                self._from_cache = True
                self._last_saved = token
                return token

        self._from_cache = False
        log.debug("Fetching new token from bearer")
        if self._receiver is None:
            self._receiver = self._bearer.wrapped.receiver()  # type: ignore  # can't be None

        token = await self._receiver.fetch(timeout=timeout, options=options)
        if token.is_empty():
            self._last_saved = None
            return token
        await self._cache.set(token)
        self._last_saved = token
        return token

    def can_retry(
        self,
        err: Exception,
        options: dict[str, str] | None = None,
    ) -> bool:
        if self._from_cache:
            self._from_cache = False
            return True

        if self._receiver is None:
            return True  # Retry if we don't have a receiver yet

        return self._receiver.can_retry(err, options)


class RenewableFileCacheBearer(ParentBearer):
    def __init__(
        self,
        bearer: ParentBearer,
        safety_margin: timedelta | float = timedelta(hours=2),
        cache_file: str | Path = Path(DEFAULT_CONFIG_DIR) / DEFAULT_CREDENTIALS_FILE,
        throttle: timedelta | float = timedelta(minutes=5),
    ) -> None:
        self._bearer = bearer
        if isinstance(safety_margin, (float, int)):
            safety_margin = timedelta(seconds=safety_margin)
        self.safety_margin: timedelta | None = safety_margin
        name = self._bearer.name
        if name is None:
            raise ValueError("Bearer must have a name for the cache.")
        self._cache = ThrottledTokenCache(
            name=name,
            cache_file=cache_file,
            throttle=throttle,
        )

    @property
    def wrapped(self) -> ParentBearer | None:
        return self._bearer

    def receiver(self) -> ParentReceiver:
        return RenewableFileCacheReceiver(self, self._cache)



from datetime import timedelta
from logging import getLogger
from pathlib import Path

from nebius.aio.token.token import Bearer as ParentBearer
from nebius.aio.token.token import Receiver as ParentReceiver
from nebius.aio.token.token import Token
from nebius.base.constants import DEFAULT_CONFIG_DIR, DEFAULT_CREDENTIALS_FILE

from .throttled_token_cache import ThrottledTokenCache

log = getLogger(__name__)


class PureFileCacheReceiver(ParentReceiver):
    def __init__(self, cache: ThrottledTokenCache) -> None:
        super().__init__()
        self._cache = cache

    async def _fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        return await self._cache.get() or Token.empty()

    def can_retry(
        self,
        err: Exception,
        options: dict[str, str] | None = None,
    ) -> bool:
        return False


class PureFileCacheBearer(ParentBearer):
    def __init__(
        self,
        name: str,
        cache_file: str | Path = Path(DEFAULT_CONFIG_DIR) / DEFAULT_CREDENTIALS_FILE,
        throttle: timedelta | float = timedelta(minutes=5),
    ) -> None:
        self._name = name
        self._cache = ThrottledTokenCache(
            name=self._name, cache_file=cache_file, throttle=throttle
        )

    @property
    def name(self) -> str:
        return self._name

    def receiver(self) -> ParentReceiver:
        return PureFileCacheReceiver(self._cache)



from logging import getLogger
from pathlib import Path

from nebius.base.error import SDKError

from .token import Bearer as ParentBearer
from .token import Receiver as ParentReceiver
from .token import Token

log = getLogger(__name__)


class NoTokenInEnvError(SDKError):
    pass


class Receiver(ParentReceiver):
    def __init__(self, token: Token) -> None:
        super().__init__()
        self._latest = token

    async def _fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        if self._latest is None:
            raise Exception("Token has to be set")
        log.debug("static token fetched")
        return self._latest

    def can_retry(
        self,
        err: Exception,
        options: dict[str, str] | None = None,
    ) -> bool:
        return False


class Bearer(ParentBearer):
    def __init__(self, file: str | Path) -> None:
        super().__init__()
        file = Path(file).expanduser()
        with open(file, "r") as f:
            token_value = f.read().strip()
        if token_value == "":
            raise SDKError("empty token file provided")
        self._tok = Token(token_value)

    def receiver(self) -> Receiver:
        return Receiver(self._tok)



from asyncio import sleep
from collections.abc import Iterable
from datetime import datetime, timedelta
from time import time
from typing import Generic, TypeVar

from grpc import CallCredentials, Compression, StatusCode

from nebius.aio.abc import ClientChannelInterface
from nebius.aio.request import DEFAULT_TIMEOUT
from nebius.base.error import SDKError
from nebius.base.protos.unset import Unset, UnsetType
from nebius.base.protos.well_known import local_timezone

from .constant_channel import Constant
from .request_status import RequestStatus

OperationPb = TypeVar("OperationPb")
T = TypeVar("T")


class Operation(Generic[OperationPb]):
    def __init__(
        self,
        source_method: str,
        channel: ClientChannelInterface,
        operation: OperationPb,
    ) -> None:
        from nebius.api.nebius.common.v1 import (
            GetOperationRequest,
            Operation,
            OperationServiceClient,
        )
        from nebius.api.nebius.common.v1alpha1 import (
            GetOperationRequest as OldGet,
        )
        from nebius.api.nebius.common.v1alpha1 import (
            Operation as Old,
        )
        from nebius.api.nebius.common.v1alpha1 import (
            OperationServiceClient as OldClient,
        )

        self._channel = channel
        _operation: OperationPb | Operation | Old = operation
        if isinstance(_operation, Operation.__PB2_CLASS__):
            _operation = Operation(_operation)
        if isinstance(_operation, Old.__PB2_CLASS__):
            _operation = Old(_operation)

        if isinstance(_operation, Operation):
            self._service: OperationServiceClient | OldClient = OperationServiceClient(
                Constant(source_method, channel)
            )
            self._get_request_obj: type[GetOperationRequest | OldGet] = (
                GetOperationRequest
            )
        elif isinstance(_operation, Old):
            self._service = OldClient(Constant(source_method, channel))
            self._get_request_obj = OldGet
        else:
            raise SDKError(f"Operation type {type(_operation)} not supported.")

        self._operation: Operation | Old = _operation

    def __repr__(self) -> str:
        return (
            f"Operation({self.id}, resource_id: {self.resource_id}, "
            f"status: {self.status()})"
        )

    def status(self) -> RequestStatus | None:
        return self._operation.status

    def done(self) -> bool:
        return self.status() is not None

    async def update(
        self,
        metadata: Iterable[tuple[str, str]] | None = None,
        timeout: float | None | UnsetType = Unset,
        credentials: CallCredentials | None = None,
        compression: Compression | None = None,
        per_retry_timeout: float | None | UnsetType = Unset,
        retries: int | None = None,
    ) -> None:
        if self.done():
            return

        req = self._service.get(
            self._get_request_obj(id=self.id),  # type: ignore
            metadata=metadata,
            timeout=timeout,
            credentials=credentials,
            compression=compression,
            per_retry_timeout=per_retry_timeout,
            retries=retries,
        )
        new_op = await req
        self._set_new_operation(new_op._operation)  # type: ignore

    def sync_wait(
        self,
        interval: float | timedelta = 1,
        metadata: Iterable[tuple[str, str]] | None = None,
        timeout: float | None = None,
        credentials: CallCredentials | None = None,
        compression: Compression | None = None,
        poll_iteration_timeout: float | None | UnsetType = Unset,
        poll_per_retry_timeout: float | None | UnsetType = Unset,
        poll_retries: int | None = None,
    ) -> None:
        run_timeout = None if timeout is None else timeout + 0.2
        return self._channel.run_sync(
            self.wait(
                interval=interval,
                metadata=metadata,
                timeout=timeout,
                credentials=credentials,
                compression=compression,
                poll_iteration_timeout=poll_iteration_timeout,
                poll_per_retry_timeout=poll_per_retry_timeout,
                poll_retries=poll_retries,
            ),
            run_timeout,
        )

    def sync_update(
        self,
        metadata: Iterable[tuple[str, str]] | None = None,
        timeout: float | None | UnsetType = Unset,
        credentials: CallCredentials | None = None,
        compression: Compression | None = None,
        per_retry_timeout: float | None | UnsetType = Unset,
        retries: int | None = None,
    ) -> None:
        run_timeout: float | None = None
        if isinstance(timeout, (int, float)):
            run_timeout = timeout + 0.2
        elif isinstance(timeout, UnsetType):
            run_timeout = DEFAULT_TIMEOUT + 0.2
        return self._channel.run_sync(
            self.update(
                metadata=metadata,
                timeout=timeout,
                credentials=credentials,
                compression=compression,
                per_retry_timeout=per_retry_timeout,
                retries=retries,
            ),
            run_timeout,
        )

    async def wait(
        self,
        interval: float | timedelta = 1,
        metadata: Iterable[tuple[str, str]] | None = None,
        timeout: float | None = None,
        credentials: CallCredentials | None = None,
        compression: Compression | None = None,
        poll_iteration_timeout: float | UnsetType | None = Unset,
        poll_per_retry_timeout: float | UnsetType | None = Unset,
        poll_retries: int | None = None,
    ) -> None:
        start = time()
        if poll_iteration_timeout is None:
            if timeout is not None:
                poll_iteration_timeout = min(5, timeout)
        if isinstance(interval, timedelta):
            interval = interval.total_seconds()
        from nebius.aio.service_error import RequestError as ServiceRequestError

        def _is_ignorable(err: Exception) -> bool:
            # TimeoutError raised locally or RequestError with DEADLINE_EXCEEDED
            if isinstance(err, TimeoutError):
                return True
            if isinstance(err, ServiceRequestError):
                try:
                    return err.status.code == StatusCode.DEADLINE_EXCEEDED
                except Exception:  # pragma: no cover - defensive
                    return False
            return False

        async def _safe_update() -> None:
            try:
                await self.update(
                    metadata=metadata,
                    timeout=poll_iteration_timeout,
                    credentials=credentials,
                    compression=compression,
                    per_retry_timeout=poll_per_retry_timeout,
                    retries=poll_retries,
                )
            except Exception as e:  # noqa: S110
                if not _is_ignorable(e):
                    raise

        if not self.done():
            await _safe_update()
        while not self.done():
            current_time = time()
            if timeout is not None and current_time > timeout + start:
                raise TimeoutError("Operation wait timeout")
            await sleep(interval)
            await _safe_update()

    def _set_new_operation(self, operation: OperationPb) -> None:
        if isinstance(operation, self._operation.__class__):
            self._operation = operation  # type: ignore
        else:
            raise SDKError(f"Operation type {type(operation)} not supported.")

    @property
    def id(self) -> str:
        return self._operation.id

    @property
    def description(self) -> str:
        return self._operation.description

    @property
    def created_at(self) -> datetime:
        ca = self._operation.created_at
        if ca is None:  # type: ignore[unused-ignore]
            return datetime.now(local_timezone)
        return ca

    @property
    def created_by(self) -> str:
        return self._operation.created_by

    @property
    def finished_at(self) -> datetime | None:
        return self._operation.finished_at

    @property
    def resource_id(self) -> str:
        return self._operation.resource_id

    def successful(self) -> bool:
        s = self.status()
        return s is not None and s.code == StatusCode.OK

    def raw(self) -> OperationPb:
        return self._operation  # type: ignore



from collections.abc import Callable, Iterable
from logging import getLogger
from typing import Any, Generic, TypeVar

from google.protobuf.message import Message as PMessage
from grpc import CallCredentials, Compression

from nebius.aio.abc import ClientChannelInterface as Channel
from nebius.aio.constant_channel import Constant
from nebius.aio.request import Request

# from nebius.api.nebius.common.v1 import Operation
from nebius.base.metadata import Metadata
from nebius.base.protos.unset import Unset, UnsetType

Req = TypeVar("Req")
Res = TypeVar("Res")


class Client:
    # __operation_type__: Message = Operation
    __service_name__: str
    __service_deprecation_details__: str | None = None

    def __init__(self, channel: Channel) -> None:
        self._channel = channel

        if self.__service_deprecation_details__ is not None:
            getLogger("deprecation").warning(
                f"Service {self.__service_name__} is deprecated. "
                f"{self.__service_deprecation_details__}",
                stack_info=True,
                stacklevel=2,
            )

    def request(
        self,
        method: str,
        request: Req,
        result_pb2_class: type[PMessage],
        metadata: Metadata | Iterable[tuple[str, str]] | None = None,
        timeout: float | None | UnsetType = Unset,
        credentials: CallCredentials | None = None,
        compression: Compression | None = None,
        auth_timeout: float | None | UnsetType = Unset,
        auth_options: dict[str, str] | None = None,
        result_wrapper: Callable[[str, Channel, Any], Res] | None = None,
        retries: int | None = 3,
        per_retry_timeout: float | None | UnsetType = Unset,
    ) -> Request[Req, Res]:
        return Request[Req, Res](
            channel=self._channel,
            service=self.__service_name__,
            method=method,
            request=request,
            metadata=metadata,
            auth_timeout=auth_timeout,
            auth_options=auth_options,
            result_pb2_class=result_pb2_class,
            timeout=timeout,
            credentials=credentials,
            compression=compression,
            result_wrapper=result_wrapper,
            retries=retries,
            per_retry_timeout=per_retry_timeout,
        )


OperationPb = TypeVar("OperationPb")
OperationService = TypeVar("OperationService", bound=Client)


class ClientWithOperations(Client, Generic[OperationPb, OperationService]):
    __operation_type__: type[OperationPb]
    __operation_service_class__: type[OperationService]
    __operation_source_method__: str

    def __init__(self, channel: Channel) -> None:
        super().__init__(channel)
        self.__operation_service__: OperationService | None = None

    def operation_service(self) -> OperationService:
        if self.__operation_service__ is None:
            self.__operation_service__ = self.__operation_service_class__(
                Constant(
                    self.__service_name__ + "." + self.__operation_source_method__,
                    self._channel,
                )
            )
        return self.__operation_service__



from abc import ABC, abstractmethod

from nebius.base.metadata import Metadata


class Authenticator(ABC):
    @abstractmethod
    async def authenticate(
        self,
        metadata: Metadata,
        timeout: float | None = None,
        options: dict[str, str] | None = None,
    ) -> None:
        raise NotImplementedError("Method not implemented!")

    @abstractmethod
    def can_retry(
        self,
        err: Exception,
        options: dict[str, str] | None = None,
    ) -> bool:
        return False


class Provider(ABC):
    @abstractmethod
    def authenticator(self) -> Authenticator:
        raise NotImplementedError("Method not implemented!")



from asyncio import (
    FIRST_COMPLETED,
    CancelledError,
    Event,
    Future,
    Task,
    create_task,
    gather,
    sleep,
    wait,
    wait_for,
)
from collections.abc import Awaitable
from datetime import datetime, timedelta, timezone
from logging import getLogger
from pathlib import Path
from typing import Any, TypeVar

from nebius.aio.token.token import Bearer as ParentBearer
from nebius.aio.token.token import Receiver as ParentReceiver
from nebius.aio.token.token import Token
from nebius.base.constants import DEFAULT_CONFIG_DIR, DEFAULT_CREDENTIALS_FILE

from ..options import (
    OPTION_MAX_RETRIES,
    OPTION_RENEW_REQUEST_TIMEOUT,
    OPTION_RENEW_REQUIRED,
    OPTION_RENEW_SYNCHRONOUS,
    OPTION_REPORT_ERROR,
)
from ..renewable import (
    RenewalError,
)
from .throttled_token_cache import ThrottledTokenCache

log = getLogger(__name__)


class AsynchronousRenewableFileCacheReceiver(ParentReceiver):
    def __init__(
        self,
        bearer: "AsynchronousRenewableFileCacheBearer",
        max_retries: int = 2,
    ) -> None:
        super().__init__()
        self._parent = bearer
        self._max_retries = max_retries
        self._trial = 0

    async def _fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        self._trial += 1
        log.debug(
            f"token fetch requested, attempt: {self._trial}," f"timeout: {timeout}"
        )
        return await self._parent.fetch(timeout=timeout, options=options)

    def can_retry(
        self,
        err: Exception,
        options: dict[str, str] | None = None,
    ) -> bool:
        max_retries = self._max_retries
        synchronous = False
        if options is not None:
            synchronous = options.get(OPTION_RENEW_SYNCHRONOUS, "") != ""
            if OPTION_MAX_RETRIES in options:
                value = options[OPTION_MAX_RETRIES]
                try:
                    max_retries = int(value)
                except ValueError as val_err:
                    log.error(
                        f"option {OPTION_MAX_RETRIES} is not valid integer: {val_err=}"
                    )
        if self._trial >= max_retries:
            log.debug("max retries reached, cannot retry")
            return False
        if not synchronous:
            self._parent.request_renewal()
        return True


VERY_LONG_TIMEOUT = timedelta(days=365 * 10)  # 10 years, should be enough for anyone

T = TypeVar("T")


class AsynchronousRenewableFileCacheBearer(ParentBearer):
    def __init__(
        self,
        source: ParentBearer,
        max_retries: int = 2,
        initial_safety_margin: timedelta | float | None = timedelta(hours=2),
        retry_safety_margin: timedelta = timedelta(hours=2),
        lifetime_safe_fraction: float = 0.9,
        initial_retry_timeout: timedelta = timedelta(seconds=1),
        max_retry_timeout: timedelta = timedelta(minutes=1),
        retry_timeout_exponent: float = 1.5,
        refresh_request_timeout: timedelta = timedelta(seconds=5),
        file_cache_throttle: timedelta | float = timedelta(minutes=5),
    ) -> None:
        super().__init__()
        self._source = source
        if isinstance(initial_safety_margin, (float, int)):
            initial_safety_margin = timedelta(seconds=initial_safety_margin)
        self._retry_safety_margin = retry_safety_margin
        self.safety_margin = initial_safety_margin
        name = self._source.name
        if name is None:
            raise ValueError("Source bearer must have a name for the cache.")
        self._file_cache = ThrottledTokenCache(
            name=name,
            cache_file=Path(DEFAULT_CONFIG_DIR) / DEFAULT_CREDENTIALS_FILE,
            throttle=file_cache_throttle,
        )

        self._is_fresh = Event()
        self._is_stopped = Event()
        self._renew_requested = Event()
        self._synchronous_can_proceed = Event()
        self._break_previous_attempt = Event()

        self._synchronous_can_proceed.set()

        self._refresh_task: Task[Any] | None = None
        self._tasks = set[Task[Any]]()

        self._renewal_attempt = 0

        self._max_retries = max_retries
        self._lifetime_safe_fraction = lifetime_safe_fraction
        self._initial_retry_timeout = initial_retry_timeout
        self._max_retry_timeout = max_retry_timeout
        self._retry_timeout_exponent = retry_timeout_exponent
        self._refresh_request_timeout = refresh_request_timeout

        self._renew_synchronous_timeout: float | None = None
        self._renewal_future: Future[Token] | None = None
        self._renew_synchronous_options: dict[str, str] | None = None

    @property
    def wrapped(self) -> ParentBearer | None:
        """Return the wrapped bearer."""
        return self._source

    def bg_task(self, coro: Awaitable[T]) -> Task[None]:
        """Run a coroutine without awaiting or tracking, and log any exceptions."""

        async def wrapper() -> None:
            try:
                await coro
            except CancelledError:
                pass
            except Exception as e:
                log.error("Unhandled exception in fire-and-forget task", exc_info=e)

        ret = create_task(wrapper())
        ret.add_done_callback(lambda x: self._tasks.discard(x))
        self._tasks.add(ret)
        return ret

    async def fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        required = False
        synchronous = False
        report_error = False
        if options is not None:
            required = options.get(OPTION_RENEW_REQUIRED, "") != ""
            synchronous = options.get(OPTION_RENEW_SYNCHRONOUS, "") != ""
            report_error = options.get(OPTION_REPORT_ERROR, "") != ""

        tok = await self._file_cache.get()
        if not required and tok is not None and not tok.is_expired():
            if self.safety_margin is None or (
                not tok.expiration
                or (tok.expiration - self.safety_margin > datetime.now(timezone.utc))
            ):
                log.debug(f"token is fresh: {tok}")
                if self._refresh_task is None:
                    log.debug("no refresh task yet, starting it")
                    self._refresh_task = self.bg_task(self._run(True))
                return tok
        self.safety_margin = None  # reset safety margin after first fetch

        if self._refresh_task is None:
            log.debug("no refresh task yet, starting it")
            self._refresh_task = self.bg_task(self._run())
        if self.is_renewal_required() or required:
            log.debug(f"renewal required, timeout {timeout}")
            if synchronous:
                self._break_previous_attempt.set()
                await wait_for(self._synchronous_can_proceed.wait(), timeout)
                if OPTION_RENEW_REQUEST_TIMEOUT in options:  # type: ignore
                    try:
                        self._renew_synchronous_timeout = float(
                            options[OPTION_RENEW_REQUEST_TIMEOUT]  # type: ignore
                        )
                    except ValueError as err:
                        log.error(
                            f"option {OPTION_RENEW_REQUEST_TIMEOUT} value is not float:"
                            f" {err=}"
                        )
                self._renew_synchronous_options = options.copy()  # type: ignore
            if report_error or synchronous:
                self._renewal_future = Future[Token]()

            self._renew_requested.set()
            if report_error or synchronous:
                return await wait_for(self._renewal_future, timeout)  # type: ignore
            else:
                await wait_for(self._is_fresh.wait(), timeout)
        tok = await self._file_cache.get()
        if tok is None:
            raise RenewalError("cache is empty after renewal")
        return tok

    async def _fetch_once(self) -> Token:
        tok = None
        log.debug(f"refreshing token, attempt {self._renewal_attempt}")
        self._break_previous_attempt.clear()
        self._synchronous_can_proceed.clear()
        timeout = self._refresh_request_timeout.total_seconds()
        if self._renew_synchronous_timeout is not None:
            timeout = self._renew_synchronous_timeout
            self._renew_synchronous_timeout = None
        token_task = create_task(self._source.receiver().fetch(timeout))
        breaker_task = create_task(self._break_previous_attempt.wait())
        _done, pending = await wait(
            [token_task, breaker_task],
            return_when=FIRST_COMPLETED,
        )
        self._renew_requested.clear()
        self._synchronous_can_proceed.set()
        for t in pending:
            t.cancel()
            self.bg_task(t)
        tok = token_task.result()
        log.debug(f"received new token: {tok}")
        if self._renewal_future is not None and not self._renewal_future.done():
            self._renewal_future.set_result(tok)
        await self._file_cache.set(tok)
        self._renewal_attempt = 0
        self._is_fresh.set()
        return tok

    async def _run(self, wait_for_timeout: bool = False) -> None:
        log.debug("refresh task started")

        if wait_for_timeout:
            tok = await self._file_cache.get()
            if tok is not None and not tok.is_expired():
                if tok.expiration is not None:
                    retry_timeout = (
                        tok.expiration - datetime.now(timezone.utc)
                    ).total_seconds() - self._retry_safety_margin.total_seconds()
                else:
                    retry_timeout = VERY_LONG_TIMEOUT.total_seconds()
            else:
                retry_timeout = 0
        else:
            retry_timeout = 0

        while not self._is_stopped.is_set():
            log.debug(
                f"Will refresh token after {retry_timeout} seconds, "
                f"renewal attempt number {self._renewal_attempt}"
            )
            _done, pending = await wait(
                [
                    self.bg_task(self._renew_requested.wait()),
                    self.bg_task(sleep(retry_timeout)),
                ],
                return_when=FIRST_COMPLETED,
            )
            for task in pending:
                task.cancel()
            await gather(*pending, return_exceptions=True)
            try:
                tok = await self._fetch_once()
                exp = tok.expiration
                if exp is None:
                    retry_timeout = VERY_LONG_TIMEOUT.total_seconds()
                else:
                    retry_timeout = (
                        exp - datetime.now(timezone.utc)
                    ).total_seconds() * self._lifetime_safe_fraction
            except Exception as e:
                log.error(
                    f"Failed refresh token, attempt: {self._renewal_attempt}, "
                    f"error: {e}",
                    exc_info=e,
                )
                if self._renewal_future is not None and not self._renewal_future.done():
                    self._renewal_future.set_exception(e)
                if (
                    self._renewal_attempt <= 1
                    or abs(self._retry_timeout_exponent - 1) < 1e-9
                ):
                    retry_timeout = self._initial_retry_timeout.total_seconds()
                else:
                    mul = self._retry_timeout_exponent ** (self._renewal_attempt - 1)
                    retry_timeout = min(
                        self._initial_retry_timeout.total_seconds() * mul,
                        self._max_retry_timeout.total_seconds(),
                    )
            if retry_timeout < self._initial_retry_timeout.total_seconds():
                retry_timeout = self._initial_retry_timeout.total_seconds()

    async def close(self, grace: float | None = None) -> None:
        source_close = create_task(self._source.close(grace=grace))
        self.stop()
        for task in self._tasks:
            task.cancel()
        rets = await gather(
            source_close,
            *self._tasks,
            return_exceptions=True,
        )
        for ret in rets:
            if isinstance(ret, BaseException) and not isinstance(ret, CancelledError):
                log.error(f"Error while graceful shutdown: {ret}", exc_info=ret)

    def is_renewal_required(self) -> bool:
        return self._file_cache.get_cached() is None or self._renew_requested.is_set()

    def request_renewal(self) -> None:
        if not self._is_stopped.is_set():
            log.debug("token renewal requested")
            self._is_fresh.clear()
            self._renew_requested.set()

    def stop(self) -> None:
        log.debug("stopping renewal task")
        self._is_stopped.set()
        self._is_fresh.clear()
        self._break_previous_attempt.set()
        self._renew_requested.set()

    def receiver(self) -> AsynchronousRenewableFileCacheReceiver:
        return AsynchronousRenewableFileCacheReceiver(
            self,
            max_retries=self._max_retries,
        )



from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Any

from nebius.base.token_sanitizer import TokenSanitizer

sanitizer = TokenSanitizer.access_token_sanitizer()


class Token:
    def __init__(self, token: str, expiration: datetime | None = None) -> None:
        self._tok = token
        self._exp = expiration

    def __str__(self) -> str:
        if self.is_empty():
            return "Token(empty)"
        ret = ["Token("]
        ret.append(sanitizer.sanitize(self._tok))
        if self._exp is not None:
            ret.append(f", expiration={self._exp.isoformat()}")
            expires_in = self._exp - datetime.now(timezone.utc)
            ret.append(f", expires_in={expires_in}")
        else:
            ret.append(", expiration=None")
        ret.append(")")
        return "".join(ret)

    @classmethod
    def empty(cls) -> "Token":
        """
        Create an empty token.
        """
        return cls(token="")

    @property
    def token(self) -> str:
        return self._tok

    @property
    def expiration(self) -> datetime | None:
        return self._exp

    def is_empty(self) -> bool:
        return self._tok == ""

    def is_expired(self) -> bool:
        if self._exp is None:
            return False
        return datetime.now(timezone.utc) >= self._exp

    def to_dict(self) -> dict[str, Any]:
        expires_at = int(self._exp.timestamp()) if self._exp is not None else 0
        data: dict[str, Any] = {
            "token": self._tok,
            "expires_at": expires_at if self._exp is not None else None,
        }
        return data

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Token):
            return NotImplemented
        return self._tok == value._tok and self._exp == value._exp

    @classmethod
    def from_dict(cls, data: Any) -> "Token":
        if not isinstance(data, dict):
            raise ValueError(
                f"Invalid format for Token: {type(data)} expected a dictionary.",
            )
        token = data.get("token", "")  # type: ignore[assignment,unused-ignore]
        if token is None:
            token = ""
        if not isinstance(token, str):
            raise ValueError(f"Invalid token format: {type(token)} expected a string.")  # type: ignore[assignment,unused-ignore]
        expires_at = data.get("expires_at", None)  # type: ignore[assignment,unused-ignore]
        if expires_at is None:
            return cls(token=token)
        if not isinstance(expires_at, int):
            raise ValueError(
                f"Invalid expires_at format: {type(expires_at)} expected an int."  # type: ignore[assignment,unused-ignore]
            )
        expiration = (
            datetime.fromtimestamp(expires_at, tz=timezone.utc) if expires_at else None
        )
        return cls(token=token, expiration=expiration)


class Receiver(ABC):
    _latest: Token | None

    @abstractmethod
    async def _fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        raise NotImplementedError("Method not implemented!")

    @property
    def latest(self) -> Token | None:
        return self._latest

    async def fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        tok = await self._fetch(timeout=timeout, options=options)
        self._latest = tok
        return tok

    @abstractmethod
    def can_retry(
        self,
        err: Exception,
        options: dict[str, str] | None = None,
    ) -> bool:
        return False


class Bearer(ABC):
    @abstractmethod
    def receiver(self) -> Receiver:
        raise NotImplementedError("Method not implemented!")

    @property
    def name(self) -> str | None:
        if self.wrapped is not None:
            return self.wrapped.name
        return None

    @property
    def wrapped(self) -> "Bearer|None":
        return None

    async def close(self, grace: float | None = None) -> None:
        if self.wrapped is not None:
            await self.wrapped.close(grace=grace)
        return None


class NamedBearer(Bearer):
    def __init__(self, wrapped: Bearer, name: str) -> None:
        self._wrapped = wrapped
        self._name = name

    @property
    def wrapped(self) -> Bearer:
        return self._wrapped

    @property
    def name(self) -> str | None:
        return self._name

    def receiver(self) -> Receiver:
        return self._wrapped.receiver()



from datetime import datetime, timedelta, timezone
from logging import getLogger
from pathlib import Path

from nebius.aio.token.token import Token
from nebius.base.constants import DEFAULT_CONFIG_DIR, DEFAULT_CREDENTIALS_FILE

from .token_cache import TokenCache

log = getLogger(__name__)


class ThrottledTokenCache:
    """
    A throttled file-based token cache.
    """

    def __init__(
        self,
        name: str,
        cache_file: str | Path = Path(DEFAULT_CONFIG_DIR) / DEFAULT_CREDENTIALS_FILE,
        throttle: timedelta | float = timedelta(minutes=5),
    ) -> None:
        self._name = name
        self._cache = TokenCache(cache_file)
        if isinstance(throttle, (float, int)):
            throttle = timedelta(seconds=throttle)
        self._throttle: timedelta = throttle
        self._cached_token: Token | None = None
        self._next_check: datetime = datetime.now(timezone.utc)

    def get_cached(self) -> Token | None:
        """
        Get the cached token without checking the throttle.
        """
        return self._cached_token

    async def get(self) -> Token | None:
        """
        Get the cached token, respecting the throttle.
        """
        if (
            self._cached_token is not None
            and not self._cached_token.is_expired()
            and self._next_check > datetime.now(timezone.utc)
        ):
            return self._cached_token

        return await self.refresh()

    async def set(self, token: Token) -> None:
        """
        Set the token in the cache and update the throttle.
        """
        if self._cached_token == token:
            return
        await self._cache.set(self._name, token)
        self._cached_token = token
        self._next_check = datetime.now(timezone.utc) + self._throttle

    async def remove(self) -> None:
        """
        Remove the token from the cache.
        """
        await self._cache.remove(self._name)
        self._cached_token = None
        self._next_check = datetime.now(timezone.utc)

    async def remove_if_equal(self, token: Token) -> None:
        """
        Remove the token from the cache if it matches the provided token.
        """
        await self._cache.remove_if_equal(self._name, token)
        if self._cached_token == token:
            self._cached_token = None
            self._next_check = datetime.now(timezone.utc)

    async def refresh(self) -> Token | None:
        """
        Refresh the cached token by fetching it again.
        """
        token = await self._cache.get(self._name)
        if token is not None and not token.is_expired():
            self._cached_token = token
            self._next_check = datetime.now(timezone.utc) + self._throttle
            return token
        return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nebius/mk8s/v1alpha1/condition.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$nebius/mk8s/v1alpha1/condition.proto\x12\x14nebius.mk8s.v1alpha1\x1a\x1fgoogle/protobuf/timestamp.proto\"\xb2\x04\n\tCondition\x12\x12\n\x04type\x18\x01 \x01(\tR\x04type\x12>\n\x06status\x18\x02 \x01(\x0e\x32&.nebius.mk8s.v1alpha1.Condition.StatusR\x06status\x12H\n\x12last_transition_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10lastTransitionAt\x12\x16\n\x06reason\x18\x04 \x01(\tR\x06reason\x12\x44\n\x08severity\x18\x05 \x01(\x0e\x32(.nebius.mk8s.v1alpha1.Condition.SeverityR\x08severity\x12 \n\x0b\x64\x65scription\x18\x06 \x01(\tR\x0b\x64\x65scription\x12\x63\n\x15last_transition_error\x18\x07 \x01(\x0b\x32/.nebius.mk8s.v1alpha1.Condition.TransitionErrorR\x13lastTransitionError\x1aK\n\x0fTransitionError\x12\x16\n\x06reason\x18\x01 \x01(\tR\x06reason\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription\")\n\x08Severity\x12\x08\n\x04NONE\x10\x00\x12\x08\n\x04INFO\x10\x01\x12\t\n\x05\x45RROR\x10\x02\"*\n\x06Status\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x08\n\x04TRUE\x10\x01\x12\t\n\x05\x46\x41LSE\x10\x02\x42\x63\n\x1b\x61i.nebius.pub.mk8s.v1alpha1B\x0e\x43onditionProtoP\x01Z2github.com/nebius/gosdk/proto/nebius/mk8s/v1alpha1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nebius.mk8s.v1alpha1.condition_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\033ai.nebius.pub.mk8s.v1alpha1B\016ConditionProtoP\001Z2github.com/nebius/gosdk/proto/nebius/mk8s/v1alpha1'
  _globals['_CONDITION']._serialized_start=96
  _globals['_CONDITION']._serialized_end=658
  _globals['_CONDITION_TRANSITIONERROR']._serialized_start=496
  _globals['_CONDITION_TRANSITIONERROR']._serialized_end=571
  _globals['_CONDITION_SEVERITY']._serialized_start=573
  _globals['_CONDITION_SEVERITY']._serialized_end=614
  _globals['_CONDITION_STATUS']._serialized_start=616
  _globals['_CONDITION_STATUS']._serialized_end=658
# @@protoc_insertion_point(module_scope)



from logging import getLogger
from os import environ
from os.path import isfile
from pathlib import Path
from ssl import SSLContext
from typing import Any, TextIO

from nebius.aio.abc import ClientChannelInterface
from nebius.aio.authorization.authorization import Provider as AuthorizationProvider
from nebius.aio.token.service_account import ServiceAccountBearer
from nebius.aio.token.static import EnvBearer, NoTokenInEnvError
from nebius.aio.token.token import Bearer as TokenBearer
from nebius.aio.token.token import Token
from nebius.base.constants import (
    DEFAULT_CONFIG_DIR,
    DEFAULT_CONFIG_FILE,
    PROFILE_ENV,
    TOKEN_ENV,
)
from nebius.base.error import SDKError
from nebius.base.service_account.service_account import (
    TokenRequester as TokenRequestReader,
)

Credentials = AuthorizationProvider | TokenBearer | TokenRequestReader | Token | str


log = getLogger(__name__)


class ConfigError(SDKError):
    pass


class NoParentIdError(ConfigError):
    pass


class Config:
    def __init__(
        self,
        client_id: str | None = None,
        config_file: str | Path = Path(DEFAULT_CONFIG_DIR) / DEFAULT_CONFIG_FILE,
        profile: str | None = None,
        profile_env: str = PROFILE_ENV,
        token_env: str = TOKEN_ENV,
        no_env: bool = False,
        no_parent_id: bool = False,
        max_retries: int = 2,
    ) -> None:
        self._client_id = client_id
        self._priority_bearer: EnvBearer | None = None
        self._profile_name = profile
        if not no_env:
            try:
                self._priority_bearer = EnvBearer(env_var_name=token_env)
            except NoTokenInEnvError:
                pass
            if self._profile_name is None:
                self._profile_name = environ.get(profile_env, None)
        self._no_parent_id = no_parent_id
        self._config_file = Path(config_file).expanduser()
        self._endpoint: str | None = None
        self._max_retries = max_retries
        self._get_profile()

    @property
    def parent_id(self) -> str:
        if self._no_parent_id:
            raise NoParentIdError(
                "Config is set to not use parent id from the profile."
            )
        if "parent-id" not in self._profile:
            raise NoParentIdError("Missing parent-id in the profile.")
        if not isinstance(self._profile["parent-id"], str):
            raise ConfigError(
                "Parent id should be a string, got "
                f"{type(self._profile['parent-id'])}."
            )
        if self._profile["parent-id"] == "":
            raise NoParentIdError("Parent id is empty.")

        return self._profile["parent-id"]

    def endpoint(self) -> str:
        return self._endpoint or ""

    def _get_profile(self) -> None:
        """Get the profile from the config file."""
        import yaml

        if not isfile(self._config_file):
            raise FileNotFoundError(f"Config file {self._config_file} not found.")

        with open(self._config_file, "r") as f:
            config = yaml.safe_load(f)

        if "profiles" not in config:
            raise ConfigError("No profiles found in the config file.")
        if not isinstance(config["profiles"], dict):
            raise ConfigError(
                f"Profiles should be a dictionary, got {type(config['profiles'])}."
            )
        if len(config["profiles"]) == 0:
            raise ConfigError(
                "No profiles found in the config file, setup the nebius CLI profile"
                " first."
            )
        if self._profile_name is None:
            if "default" not in config:
                if len(config["profiles"]) == 1:
                    self._profile_name = list(config["profiles"].keys())[0]
                else:
                    raise ConfigError("No default profile found in the config file.")
            else:
                self._profile_name = config["default"]
            if self._profile_name is None:
                raise ConfigError(
                    "No profile selected. Either set the profile in the config setup,"
                    " set the env var NEBIUS_PROFILE or "
                    "execute `nebius profile activate`."
                )
        profile = self._profile_name
        if not isinstance(profile, str):
            raise ConfigError(f"Profile name should be a string, got {type(profile)}.")
        if profile not in config["profiles"]:
            raise ConfigError(f"Profile {profile} not found in the config file.")
        if not isinstance(config["profiles"][profile], dict):
            raise ConfigError(
                f"Profile {profile} should be a dictionary, got "
                f"{type(config['profiles'][profile])}."
            )
        self._profile: dict[str, Any] = config["profiles"][profile]

        if "endpoint" in self._profile:
            if not isinstance(self._profile["endpoint"], str):
                raise ConfigError(
                    "Endpoint should be a string, got "
                    f"{type(self._profile['endpoint'])}."
                )
            self._endpoint = self._profile["endpoint"]

    def get_credentials(
        self,
        channel: ClientChannelInterface,
        writer: TextIO | None = None,
        no_browser_open: bool = False,
        ssl_ctx: SSLContext | None = None,
    ) -> Credentials:
        if self._priority_bearer is not None:
            return self._priority_bearer
        if "token-file" in self._profile:
            from nebius.aio.token.file import Bearer as FileBearer

            if not isinstance(self._profile["token-file"], str):
                raise ConfigError(
                    "Token file should be a string, got "
                    f" {type(self._profile['token-file'])}."
                )
            return FileBearer(self._profile["token-file"])
        if "auth-type" not in self._profile:
            raise ConfigError("Missing auth-type in the profile.")
        auth_type = self._profile["auth-type"]
        if auth_type == "federation":
            if "federation-endpoint" not in self._profile:
                raise ConfigError("Missing federation-endpoint in the profile.")
            if not isinstance(self._profile["federation-endpoint"], str):
                raise ConfigError(
                    "Federation endpoint should be a string, got "
                    f"{type(self._profile['federation-endpoint'])}."
                )
            if "federation-id" not in self._profile:
                raise ConfigError("Missing federation-id in the profile.")
            if not isinstance(self._profile["federation-id"], str):
                raise ConfigError(
                    "Federation id should be a string, got "
                    f"{type(self._profile['federation-id'])}."
                )
            from nebius.aio.token.federation_account import FederationBearer

            if not self._client_id:
                raise ConfigError(
                    "Client ID is required for federation authentication."
                )

            log.debug(
                f"Creating FederationBearer with profile {self._profile_name}, "
                f"client_id {self._client_id}, "
                f"federation_url {self._profile['federation-endpoint']}, "
                f"federation_id {self._profile['federation-id']}, "
                f"writer {writer}, no_browser_open {no_browser_open}."
            )

            # Federation bearer already wraps underlying signer with a file-cache
            # renewable bearer. Return it directly.
            return FederationBearer(
                profile_name=self._profile_name,  # type: ignore
                client_id=self._client_id,
                federation_endpoint=self._profile["federation-endpoint"],
                federation_id=self._profile["federation-id"],
                writer=writer,
                no_browser_open=no_browser_open,
                ssl_ctx=ssl_ctx,
            )
        elif auth_type == "service account":
            from cryptography.hazmat.backends import default_backend
            from cryptography.hazmat.primitives import serialization
            from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

            # Possible sources (priority):
            # 1) federated-subject-credentials-file-path + service-account-id
            # 2) service-account-credentials-file-path
            # 3) inline private-key with service-account-id + public-key-id
            # 4) private-key-file-path with service-account-id + public-key-id

            svc_id: str | None = None
            if "service-account-id" in self._profile:
                if not isinstance(self._profile["service-account-id"], str):
                    raise ConfigError(
                        "Service account should be a string, got "
                        f"{type(self._profile['service-account-id'])}."
                    )
                svc_id = self._profile["service-account-id"]

            # 1) federated subject credentials file
            if (
                svc_id is not None
                and "federated-subject-credentials-file-path" in self._profile
            ):
                if not isinstance(
                    self._profile["federated-subject-credentials-file-path"], str
                ):
                    raise ConfigError(
                        "federated-subject-credentials-file-path should be a string"
                    )
                # Use the federated credentials bearer (file-backed) to supply
                # the actor token for token-exchange. Do not treat this as a
                # service-account credentials file.
                from nebius.aio.token.federated_credentials import (
                    FederatedCredentialsBearer,
                )

                return FederatedCredentialsBearer(
                    self._profile["federated-subject-credentials-file-path"],
                    service_account_id=svc_id,
                    channel=channel,
                )

            # 2) service account credentials file
            if "service-account-credentials-file-path" in self._profile:
                if not isinstance(
                    self._profile["service-account-credentials-file-path"], str
                ):
                    raise ConfigError(
                        "service-account-credentials-file-path should be a string"
                    )
                from nebius.base.service_account.credentials_file import (
                    Reader as CredentialsFileReader,
                )

                return ServiceAccountBearer(
                    service_account=CredentialsFileReader(
                        self._profile["service-account-credentials-file-path"]
                    ),
                    channel=channel,
                )

            # 3 & 4) inline private-key or private-key file path
            if svc_id is None:
                raise ConfigError("Missing service-account-id in the profile.")

            if "public-key-id" not in self._profile:
                raise ConfigError("Missing public-key-id in the profile.")
            if not isinstance(self._profile["public-key-id"], str):
                raise ConfigError(
                    "Public key should be a string, got "
                    f"{type(self._profile['public-key-id'])}."
                )
            pk_id = self._profile["public-key-id"]

            # inline private-key
            if "private-key" in self._profile:
                if not isinstance(self._profile["private-key"], str):
                    raise ConfigError(
                        "Private key should be a string, got "
                        f"{type(self._profile['private-key'])}."
                    )
                pk = serialization.load_pem_private_key(
                    self._profile["private-key"].encode("utf-8"),
                    password=None,
                    backend=default_backend(),
                )
                if not isinstance(pk, RSAPrivateKey):
                    raise ConfigError(
                        f"Private key should be of type RSAPrivateKey, got {type(pk)}."
                    )
                return ServiceAccountBearer(
                    service_account=svc_id,
                    public_key_id=pk_id,
                    private_key=pk,
                    channel=channel,
                )

            # private-key file path
            if "private-key-file-path" in self._profile:
                if not isinstance(self._profile["private-key-file-path"], str):
                    raise ConfigError("private-key-file-path should be a string")
                from nebius.base.service_account.pk_file import Reader as PKFileReader

                return ServiceAccountBearer(
                    service_account=PKFileReader(
                        self._profile["private-key-file-path"], pk_id, svc_id
                    ),
                    channel=channel,
                )

            # Nothing matched
            raise ConfigError(
                "Incomplete service account configuration: provide either "
                "(service-account-id and federated-subject-credentials-file-path) OR "
                "(service-account-credentials-file-path) OR "
                "(service-account-id, public-key-id and one of "
                "private-key / private-key-file-path)"
            )
        else:
            raise ConfigError(f"Unsupported auth-type {auth_type} in the profile.")



import os
from logging import getLogger

from nebius.base.constants import TOKEN_ENV
from nebius.base.error import SDKError

from .token import Bearer as ParentBearer
from .token import Receiver as ParentReceiver
from .token import Token

log = getLogger(__name__)


class NoTokenInEnvError(SDKError):
    pass


class Receiver(ParentReceiver):
    def __init__(self, token: Token) -> None:
        super().__init__()
        self._latest = token

    async def _fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        if self._latest is None:
            raise Exception("Token has to be set")
        log.debug("static token fetched")
        return self._latest

    def can_retry(
        self,
        err: Exception,
        options: dict[str, str] | None = None,
    ) -> bool:
        return False


class Bearer(ParentBearer):
    def __init__(self, token: Token | str) -> None:
        if isinstance(token, str):
            token = Token(token)
        if token.token == "":
            raise SDKError("empty token provided")
        super().__init__()
        self._tok = token

    def receiver(self) -> Receiver:
        return Receiver(self._tok)


class EnvBearer(Bearer):
    def __init__(self, env_var_name: str = TOKEN_ENV) -> None:
        val = os.environ.get(env_var_name, "")
        if val == "":
            raise NoTokenInEnvError(f"no token in env {env_var_name}")
        super().__init__(val)



from collections.abc import Callable
from logging import getLogger
from typing import TypeVar
from uuid import uuid4

from grpc.aio._call import UnaryUnaryCall
from grpc.aio._interceptor import ClientCallDetails, UnaryUnaryClientInterceptor
from grpc.aio._metadata import Metadata as GRPCMetadata

from nebius.base.metadata import Metadata

log = getLogger(__name__)

HEADER = "x-idempotency-key"

Req = TypeVar("Req")
Res = TypeVar("Res")


def new_key() -> str:
    return str(uuid4())


def add_key_to_metadata(metadata: Metadata | GRPCMetadata) -> None:
    log.debug("added idempotency key to metadata")
    metadata[HEADER] = new_key()


def ensure_key_in_metadata(metadata: Metadata | GRPCMetadata) -> None:
    if HEADER not in metadata or metadata[HEADER] == "" or metadata[HEADER] == [""]:
        add_key_to_metadata(metadata)


class IdempotencyKeyInterceptor(UnaryUnaryClientInterceptor):  # type: ignore[unused-ignore,misc]
    async def intercept_unary_unary(
        self,
        continuation: Callable[[ClientCallDetails, Req], UnaryUnaryCall | Res],
        client_call_details: ClientCallDetails,
        request: Req,
    ) -> UnaryUnaryCall | Res:
        if client_call_details.metadata is None:
            client_call_details.metadata = GRPCMetadata()
        ensure_key_in_metadata(client_call_details.metadata)
        return await continuation(client_call_details, request)  # type: ignore



from dataclasses import dataclass
from enum import Enum

from google.protobuf.any_pb2 import Any as AnyPb
from google.rpc.status_pb2 import Status as StatusPb  # type: ignore
from grpc import StatusCode


class UnfinishedRequestStatus(Enum):
    INITIALIZED = 1
    SENT = 2


@dataclass
class RequestStatus:
    code: StatusCode
    message: str | None
    details: list[AnyPb]
    request_id: str
    trace_id: str

    def to_rpc_status(self) -> StatusPb:  # type: ignore[unused-ignore]
        ret = StatusPb()  # type: ignore[unused-ignore]
        ret.code = self.code
        ret.message = self.message
        ret.details.extend(self.details)  # type: ignore[unused-ignore]
        return ret  # type: ignore[unused-ignore]

    @classmethod
    def from_rpc_status(
        cls,
        status: StatusPb,  # type: ignore[unused-ignore]
        request_id: str,
        trace_id: str,
    ) -> "RequestStatus":
        return cls(
            code=status.code,  # type: ignore[unused-ignore]
            message=status.message,  # type: ignore[unused-ignore]
            details=[d for d in status.details],  # type: ignore[unused-ignore]
            request_id=request_id,
            trace_id=trace_id,
        )


def request_status_from_rpc_status(status: StatusPb) -> RequestStatus:  # type: ignore[unused-ignore]
    from .service_error import RequestStatusExtended

    return RequestStatusExtended.from_rpc_status(status, request_id="", trace_id="")  # type: ignore[unused-ignore]


def request_status_to_rpc_status(status: RequestStatus) -> StatusPb:  # type: ignore[unused-ignore]
    return status.to_rpc_status()  # type: ignore[unused-ignore]



from logging import getLogger
from pathlib import Path

import yaml

from nebius.aio.token.token import Token
from nebius.base.constants import DEFAULT_CONFIG_DIR, DEFAULT_CREDENTIALS_FILE

from .async_flock import Lock

log = getLogger(__name__)


class TokenCache:
    """
    A simple file-based token cache.
    """

    def __init__(
        self,
        cache_file: str | Path = Path(DEFAULT_CONFIG_DIR) / DEFAULT_CREDENTIALS_FILE,
        path_create_mode: int = 0o750,
        file_create_mode: int = 0o600,
        flock_timeout: float | None = 5.0,
    ) -> None:
        self.cache_file = Path(cache_file).expanduser()
        self.flock_timeout = flock_timeout
        self.file_create_mode = file_create_mode
        self.path_create_mode = path_create_mode

    def _yaml_parse(self, data: str) -> dict[str, Token]:
        """
        Parse YAML data from a string.
        """
        try:
            data = yaml.safe_load(data) or {}  # type: ignore
            if not isinstance(data, dict):
                raise ValueError(
                    f"Invalid YAML format: {type(data)} expected a dictionary."
                )
            tokens_strs = data.get("tokens", {})  # type: ignore[unused-ignore]
            if not isinstance(tokens_strs, dict):
                raise ValueError(
                    f"Invalid tokens format: {type(tokens_strs)} expected a dictionary."  # type: ignore[unused-ignore]
                )
            tokens = dict[str, Token]()
            for k, v in tokens_strs.items():  # type: ignore[unused-ignore]
                if not isinstance(k, str):
                    raise ValueError(
                        f"Invalid token format: key '{k}' must be a string."
                    )

                tokens[k] = Token.from_dict(v)
            return tokens
        except yaml.YAMLError as e:
            raise ValueError(f"Invalid YAML data: {e}")

    async def get(self, name: str) -> Token | None:
        """
        Get a token from the cache by name.
        """
        try:
            if not self.cache_file.is_file():
                return None
            async with Lock(
                self.cache_file,
                "r",
                create_mode=self.file_create_mode,
                timeout=self.flock_timeout,
            ) as f:  # type: ignore[unused-ignore]
                tokens = self._yaml_parse(f.read())  # type: ignore[unused-ignore]
            ret = tokens.get(name, None)
            if ret is not None and not ret.is_expired():
                return ret
            await self.remove(name)  # Clean up expired token
            return None
        except ValueError as e:
            log.warning(
                f"Failed to parse tokens from {self.cache_file}. "
                f"Returning None for the requested token: {e}"
            )
            return None
        except FileNotFoundError:
            return None

    def _yaml_dump(self, tokens: dict[str, Token]) -> str:
        """
        Dump tokens to a YAML string.
        """
        toks = {k: v.to_dict() for k, v in tokens.items() if not v.is_expired()}
        if len(toks) == 0:
            return ""
        return yaml.dump(
            {"tokens": toks},
            sort_keys=False,
        )

    async def set(self, name: str, token: Token) -> None:
        """
        Set a token in the cache by name.
        """
        try:
            if not self.cache_file.parent.is_dir():
                self.cache_file.parent.mkdir(
                    mode=self.path_create_mode, parents=True, exist_ok=True
                )
            async with Lock(
                self.cache_file,
                "a+",
                create_mode=self.file_create_mode,
                timeout=self.flock_timeout,
            ) as f:  # type: ignore[unused-ignore]
                f.seek(0)
                try:
                    tokens = self._yaml_parse(f.read())  # type: ignore[unused-ignore]
                except ValueError as e:
                    log.warning(
                        f"Failed to parse tokens from {self.cache_file}. "
                        f"Starting with an empty cache: {e}"
                    )
                    tokens = {}
                tokens[name] = token
                f.seek(0)
                f.truncate()
                f.write(self._yaml_dump(tokens))  # type: ignore[unused-ignore]
        except ValueError as e:
            raise ValueError(f"Failed to set token: {e}")

    async def remove(self, name: str) -> None:
        """
        Remove a token from the cache by name.
        """
        try:
            if not self.cache_file.is_file():
                return
            async with Lock(
                self.cache_file,
                "r+",
                create_mode=self.file_create_mode,
                timeout=self.flock_timeout,
            ) as f:  # type: ignore[unused-ignore]
                try:
                    tokens = self._yaml_parse(f.read())  # type: ignore[unused-ignore]
                except ValueError as e:
                    log.warning(
                        f"Failed to parse tokens from {self.cache_file}. "
                        f"Starting with an empty cache: {e}"
                    )
                    tokens = {}
                if name in tokens:
                    del tokens[name]
                f.seek(0)
                f.truncate()
                f.write(self._yaml_dump(tokens))  # type: ignore[unused-ignore]
        except ValueError as e:
            raise ValueError(f"Failed to remove token: {e}")

    async def remove_if_equal(self, name: str, token: Token) -> None:
        """
        Remove a token from the cache by name if it matches the provided token.
        """
        try:
            if not self.cache_file.is_file():
                return
            async with Lock(
                self.cache_file,
                "r+",
                create_mode=self.file_create_mode,
                timeout=self.flock_timeout,
            ) as f:  # type: ignore[unused-ignore]
                try:
                    tokens = self._yaml_parse(f.read())  # type: ignore[unused-ignore]
                except ValueError as e:
                    log.warning(
                        f"Failed to parse tokens from {self.cache_file}. "
                        f"Starting with an empty cache: {e}"
                    )
                    tokens = {}
                if name in tokens and tokens[name] == token:
                    del tokens[name]
                f.seek(0)
                f.truncate()
                f.write(self._yaml_dump(tokens))  # type: ignore[unused-ignore]
        except ValueError as e:
            raise ValueError(f"Failed to remove token: {e}")



from datetime import timedelta

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from nebius.aio.abc import ClientChannelInterface
from nebius.aio.token.deferred_channel import DeferredChannel
from nebius.aio.token.exchangeable import Bearer as ExchangeableBearer
from nebius.aio.token.renewable import Bearer as RenewableBearer
from nebius.aio.token.token import Bearer as ParentBearer
from nebius.aio.token.token import NamedBearer, Receiver
from nebius.base.service_account.service_account import (
    Reader as ServiceAccountReader,
)
from nebius.base.service_account.service_account import (
    ServiceAccount,
)
from nebius.base.service_account.static import Reader as ServiceAccountReaderStatic


class ServiceAccountBearer(ParentBearer):
    def __init__(
        self,
        service_account: ServiceAccountReader | ServiceAccount | str,
        channel: ClientChannelInterface | DeferredChannel | None = None,
        private_key: RSAPrivateKey | None = None,
        public_key_id: str | None = None,
        max_retries: int = 2,
        lifetime_safe_fraction: float = 0.9,
        initial_retry_timeout: timedelta = timedelta(seconds=1),
        max_retry_timeout: timedelta = timedelta(minutes=1),
        retry_timeout_exponent: float = 1.5,
        refresh_request_timeout: timedelta = timedelta(seconds=5),
    ) -> None:
        reader: ServiceAccountReader | None = None
        if isinstance(service_account, ServiceAccountReader):
            reader = service_account
            service_account = service_account.read()
        if isinstance(service_account, str):
            if not isinstance(private_key, RSAPrivateKey):
                raise TypeError(
                    "Private key must be provided as RSAPrivateKey instance "
                    "when service_account is a string."
                )
            if not isinstance(public_key_id, str):
                raise TypeError(
                    "Public key ID must be provided as a string when service_account "
                    "is a string."
                )
            service_account = ServiceAccount(
                private_key=private_key,
                public_key_id=public_key_id,
                service_account_id=service_account,
            )
        else:
            if private_key is not None or public_key_id is not None:
                raise ValueError(
                    "Private key and public key ID must not be provided "
                    "when service_account is a ServiceAccount or ServiceAccountReader "
                    "instance."
                )
        if not isinstance(service_account, ServiceAccount):  # type: ignore[unused-ignore]
            raise TypeError(
                "service_account must be ServiceAccountReader, ServiceAccount or string"
                f", got {type(service_account)}"
            )
        if reader is None:
            reader = ServiceAccountReaderStatic(service_account)
        sa_id = service_account.service_account_id
        public_key_id = service_account.public_key_id
        private_key = service_account.private_key

        self._exchangeable = ExchangeableBearer(
            reader,
            channel=channel,
            max_retries=max_retries,
        )

        self._source = NamedBearer(
            RenewableBearer(
                self._exchangeable,
                max_retries=max_retries,
                lifetime_safe_fraction=lifetime_safe_fraction,
                initial_retry_timeout=initial_retry_timeout,
                max_retry_timeout=max_retry_timeout,
                retry_timeout_exponent=retry_timeout_exponent,
                refresh_request_timeout=refresh_request_timeout,
            ),
            f"service-account/{sa_id}/{public_key_id}",
        )

    def set_channel(self, channel: ClientChannelInterface) -> None:
        self._exchangeable.set_channel(channel)

    @property
    def wrapped(self) -> "ParentBearer|None":
        return self._source

    def receiver(self) -> "Receiver":
        return self._source.receiver()



from collections.abc import Awaitable
from typing import Protocol, TypeVar, runtime_checkable

from .authorization.authorization import Provider as AuthorizationProvider
from .base import AddressChannel

T = TypeVar("T")


class SyncronizerInterface(Protocol):
    def run_sync(self, awaitable: Awaitable[T], timeout: float | None = None) -> T: ...


@runtime_checkable
class ClientChannelInterface(Protocol):
    def get_channel_by_method(self, method_name: str) -> AddressChannel: ...
    def return_channel(self, chan: AddressChannel | None) -> None: ...
    def discard_channel(self, chan: AddressChannel | None) -> None: ...
    def get_authorization_provider(self) -> AuthorizationProvider | None: ...

    def parent_id(self) -> str | None: ...

    def run_sync(self, awaitable: Awaitable[T], timeout: float | None = None) -> T: ...


class GracefulInterface(Protocol):
    async def close(self, grace: float | None = None) -> None: ...



import sys
from asyncio import (
    FIRST_COMPLETED,
    AbstractEventLoop,
    CancelledError,
    Task,
    create_task,
    gather,
    get_event_loop,
    iscoroutine,
    new_event_loop,
    run_coroutine_threadsafe,
    sleep,
    wait,
)
from collections.abc import Awaitable, Coroutine, Sequence
from inspect import isawaitable
from logging import getLogger
from pathlib import Path
from typing import Any, TextIO, TypeVar

from google.protobuf.message import Message
from grpc import (
    CallCredentials,
    ChannelConnectivity,
    ChannelCredentials,
    Compression,
    ssl_channel_credentials,
)
from grpc.aio import Channel as GRPCChannel
from grpc.aio._base_call import UnaryUnaryCall
from grpc.aio._base_channel import (
    StreamStreamMultiCallable,
    StreamUnaryMultiCallable,
    UnaryStreamMultiCallable,
    UnaryUnaryMultiCallable,
)
from grpc.aio._channel import (
    insecure_channel,  # type: ignore[unused-ignore]
    secure_channel,  # type: ignore[unused-ignore]
)
from grpc.aio._interceptor import ClientInterceptor
from grpc.aio._typing import (
    ChannelArgumentType,
    DeserializingFunction,
    MetadataType,
    SerializingFunction,
)

from nebius.aio.abc import GracefulInterface
from nebius.aio.authorization.authorization import Provider as AuthorizationProvider
from nebius.aio.authorization.token import TokenProvider
from nebius.aio.cli_config import Config as ConfigReader
from nebius.aio.idempotency import IdempotencyKeyInterceptor
from nebius.aio.service_descriptor import ServiceStub, from_stub_class
from nebius.aio.token import exchangeable, renewable
from nebius.aio.token.static import Bearer as StaticTokenBearer
from nebius.aio.token.static import EnvBearer
from nebius.aio.token.token import Bearer as TokenBearer
from nebius.aio.token.token import Token
from nebius.api.nebius.common.v1.operation_service_pb2_grpc import (
    OperationServiceStub,
)
from nebius.api.nebius.common.v1alpha1.operation_service_pb2_grpc import (
    OperationServiceStub as OperationServiceStubDeprecated,
)
from nebius.base.constants import DOMAIN
from nebius.base.error import SDKError
from nebius.base.methods import service_from_method_name
from nebius.base.options import COMPRESSION, INSECURE, pop_option
from nebius.base.resolver import Chain, Conventional, Resolver, TemplateExpander
from nebius.base.service_account.service_account import (
    Reader as ServiceAccountReader,
)
from nebius.base.service_account.service_account import (
    TokenRequester as TokenRequestReader,
)
from nebius.base.tls_certificates import get_system_certificates
from nebius.base.version import version

from .base import AddressChannel, ChannelBase

logger = getLogger(__name__)

Req = TypeVar("Req", bound=Message)
Res = TypeVar("Res", bound=Message)

T = TypeVar("T")


class LoopError(SDKError):
    pass


class ChannelClosedError(SDKError):
    pass


class NebiusUnaryUnaryMultiCallable(UnaryUnaryMultiCallable[Req, Res]):  # type: ignore[unused-ignore,misc]
    def __init__(
        self,
        channel: "Channel",
        method: str,
        request_serializer: SerializingFunction | None = None,
        response_deserializer: DeserializingFunction | None = None,
    ) -> None:
        super().__init__()
        self._channel = channel
        self._method = method
        self._request_serializer = request_serializer
        self._response_deserializer = response_deserializer

    def __call__(
        self,
        request: Req,
        *,
        timeout: float | None = None,
        metadata: MetadataType | None = None,
        credentials: CallCredentials | None = None,
        wait_for_ready: bool | None = None,
        compression: Compression | None = None,
    ) -> UnaryUnaryCall[Req, Res]:
        ch = self._channel.get_channel_by_method(self._method)
        ret = ch.channel.unary_unary(  # type: ignore[unused-ignore,call-arg,assignment,misc]
            self._method,
            self._request_serializer,
            self._response_deserializer,
        )(  # type: ignore[unused-ignore,call-arg,assignment,misc]
            request,
            timeout=timeout,
            metadata=metadata,  # type: ignore
            credentials=credentials,
            wait_for_ready=wait_for_ready,
            compression=compression,
        )

        def return_channel(cb_arg: Any) -> None:
            nonlocal ch
            logger.debug(f"Done with call {self=}, {cb_arg=}")
            try:
                self._channel.discard_channel(ch)
            except ChannelClosedError:
                self._channel.run_sync(ch.channel.close(None), 0.1)
                # pass
            ch = None  # type: ignore

        ret.add_done_callback(return_channel)
        return ret  # type: ignore[unused-ignore,return-value]


class NoCredentials:
    pass


Credentials = (
    AuthorizationProvider
    | TokenBearer
    | TokenRequestReader
    | NoCredentials
    | Token
    | str
    | None
)


def _wrap_awaitable(awaitable: Awaitable[T]) -> Coroutine[Any, Any, T]:
    if iscoroutine(awaitable):
        return awaitable
    if not isawaitable(awaitable):
        raise TypeError(
            "An asyncio.Future, a coroutine or an awaitable is "
            + f"required, {type(awaitable)} given"
        )

    async def wrap() -> T:
        return await awaitable

    return wrap()


async def _run_awaitable_with_timeout(
    f: Awaitable[T],
    timeout: float | None = None,
) -> T:
    task = Task(_wrap_awaitable(f), name=f"Task for {f=}")
    tasks: list[Task[Any]] = list[Task[Any]]([task])
    if timeout is not None:
        timer = Task(sleep(timeout), name=f"Timer for {f=}")
        tasks.append(timer)
    done, pending = await wait(
        tasks,
        return_when=FIRST_COMPLETED,
    )
    for p in pending:
        logger.debug(f"Canceling pending task {p}")
        p.cancel()
    await gather(*pending, return_exceptions=True)
    try:
        if task.exception() is not None:
            if task not in done:
                raise TimeoutError("Awaitable timed out") from task.exception()
            raise task.exception()  # type: ignore
    except CancelledError as e:
        if task not in done:
            raise TimeoutError("Awaitable timed out") from e
        raise e
    return task.result()


def set_user_agent_option(
    user_agent: str, options: ChannelArgumentType | None
) -> ChannelArgumentType:
    options = list(options or [])
    options.append(("grpc.primary_user_agent", user_agent))
    return options


class Channel(ChannelBase):  # type: ignore[unused-ignore,misc]
    def __init__(
        self,
        *,
        resolver: Resolver | None = None,
        substitutions: dict[str, str] | None = None,
        user_agent_prefix: str | None = None,
        domain: str | None = None,
        options: ChannelArgumentType | None = None,
        interceptors: Sequence[ClientInterceptor] | None = None,
        address_options: dict[str, ChannelArgumentType] | None = None,
        address_interceptors: dict[str, Sequence[ClientInterceptor]] | None = None,
        credentials: Credentials = None,
        service_account_id: str | None = None,
        service_account_public_key_id: str | None = None,
        service_account_private_key_file_name: str | Path | None = None,
        credentials_file_name: str | Path | None = None,
        config_reader: ConfigReader | None = None,
        tls_credentials: ChannelCredentials | None = None,
        event_loop: AbstractEventLoop | None = None,
        max_free_channels_per_address: int = 2,
        parent_id: str | None = None,
        federation_invitation_writer: TextIO | None = None,
        federation_invitation_no_browser_open: bool = False,
    ) -> None:
        import nebius.api.nebius.iam.v1.token_exchange_service_pb2  # type: ignore[unused-ignore] # noqa: F401 - load for registration
        import nebius.api.nebius.iam.v1.token_exchange_service_pb2_grpc  # type: ignore[unused-ignore] # noqa: F401 - load for registration

        if domain is None:
            if config_reader is not None:
                domain = config_reader.endpoint()

            if domain is None or domain == "":
                domain = DOMAIN

        substitutions_full = dict[str, str]()
        substitutions_full["{domain}"] = domain
        if substitutions is not None:
            substitutions_full.update(substitutions)

        self._max_free_channels_per_address = max_free_channels_per_address

        self._gracefuls = set[GracefulInterface]()
        self._tasks = set[Task[Any]]()

        self._resolver: Resolver = Conventional()
        if resolver is not None:
            self._resolver = Chain(resolver, self._resolver)
        self._resolver = TemplateExpander(substitutions_full, self._resolver)
        if tls_credentials is None:
            root_ca = get_system_certificates()
            with open(root_ca, "rb") as f:
                trusted_certs = f.read()
            tls_credentials = ssl_channel_credentials(root_certificates=trusted_certs)
        self._tls_credentials = tls_credentials

        self._free_channels = dict[str, list[GRPCChannel]]()
        self._methods = dict[str, str]()
        self.user_agent = "nebius-python-sdk/" + version
        self.user_agent += f" (python/{sys.version_info.major}.{sys.version_info.minor}"
        self.user_agent += f".{sys.version_info.micro})"

        if user_agent_prefix is not None:
            self.user_agent = f"{user_agent_prefix} {self.user_agent}"

        if interceptors is None:
            interceptors = []
        self._global_options = options or []
        self._global_interceptors: list[ClientInterceptor] = [
            IdempotencyKeyInterceptor()
        ]
        self._global_interceptors.extend(interceptors)

        if address_options is None:
            address_options = dict[str, ChannelArgumentType]()
        if address_interceptors is None:
            address_interceptors = dict[str, Sequence[ClientInterceptor]]()
        self._address_options = address_options
        self._address_interceptors = address_interceptors

        self._global_interceptors_inner: list[ClientInterceptor] = []

        self._parent_id = parent_id
        if self._parent_id is None and config_reader is not None:
            from .cli_config import NoParentIdError

            try:
                self._parent_id = config_reader.parent_id
            except NoParentIdError:
                pass
        if self._parent_id == "":
            raise SDKError("Parent id is empty")

        self._token_bearer: TokenBearer | None = None
        self._authorization_provider: AuthorizationProvider | None = None
        if credentials is None:
            if credentials_file_name is not None:
                from nebius.base.service_account.credentials_file import (
                    Reader as CredentialsFileReader,
                )

                credentials = CredentialsFileReader(credentials_file_name)
            elif (
                service_account_id is not None
                and service_account_private_key_file_name is not None
                and service_account_public_key_id is not None
            ):
                from nebius.base.service_account.pk_file import Reader as PKFileReader

                credentials = PKFileReader(
                    service_account_private_key_file_name,
                    service_account_public_key_id,
                    service_account_id,
                )
            elif config_reader is not None:
                credentials = config_reader.get_credentials(
                    self,
                    writer=federation_invitation_writer,
                    no_browser_open=federation_invitation_no_browser_open,
                )
            else:
                credentials = EnvBearer()
        if isinstance(credentials, str) or isinstance(credentials, Token):
            credentials = StaticTokenBearer(credentials)
        if isinstance(credentials, ServiceAccountReader):
            from nebius.aio.token.service_account import ServiceAccountBearer

            credentials = ServiceAccountBearer(
                credentials,
                self,
            )
        if isinstance(credentials, TokenRequestReader):
            exchange = exchangeable.Bearer(credentials, self)
            cache = renewable.Bearer(exchange)
            credentials = cache
        if isinstance(credentials, TokenBearer):
            self._gracefuls.add(credentials)
            self._token_bearer = credentials
            credentials = TokenProvider(credentials)
        if isinstance(credentials, AuthorizationProvider):
            self._authorization_provider = credentials
        elif not isinstance(credentials, NoCredentials):  # type: ignore[unused-ignore]
            raise SDKError(f"credentials type is not supported: {type(credentials)}")

        self._event_loop = event_loop
        self._closed = False

    def get_authorization_provider(self) -> AuthorizationProvider | None:
        return self._authorization_provider

    async def get_token(
        self,
        timeout: float | None,
        options: dict[str, str] | None = None,
    ) -> Token:
        if self._token_bearer is None:
            raise SDKError("Token bearer is not set")
        receiver = self._token_bearer.receiver()
        return await receiver.fetch(
            timeout=timeout,
            options=options,
        )

    def get_token_sync(
        self,
        timeout: float | None,
        options: dict[str, str] | None = None,
    ) -> Token:
        timeout_sync = timeout
        if timeout_sync is not None:
            timeout_sync += 0.2  # 200 ms for graceful shutdown
        return self.run_sync(
            self.get_token(timeout, options),
            timeout_sync,
        )

    def parent_id(self) -> str | None:
        return self._parent_id

    def bg_task(self, coro: Awaitable[T]) -> Task[None]:
        """Run a coroutine without awaiting or tracking, and log any exceptions."""

        async def wrapper() -> None:
            try:
                await coro
            except CancelledError:
                pass
            except Exception as e:
                logger.error("Unhandled exception in Channel.bg_task", exc_info=e)

        ret = create_task(wrapper(), name=f"Channel.bg_task for {coro}")
        ret.add_done_callback(lambda x: self._tasks.discard(x))
        self._tasks.add(ret)
        return ret

    def run_sync(self, awaitable: Awaitable[T], timeout: float | None = None) -> T:
        loop_provided = self._event_loop is not None
        if self._event_loop is None:
            try:
                self._event_loop = get_event_loop()
            except RuntimeError:
                self._event_loop = new_event_loop()

        if self._event_loop.is_running():
            if loop_provided:
                try:
                    if get_event_loop() == self._event_loop:
                        raise LoopError(
                            "Provided loop is equal to current thread's "
                            "loop. Either use async/await or provide "
                            "another loop at the SDK initialization."
                        )
                except RuntimeError:
                    pass
                return run_coroutine_threadsafe(
                    _run_awaitable_with_timeout(awaitable, timeout),
                    self._event_loop,
                ).result()
            else:
                raise LoopError(
                    "Synchronous call inside async context. Either use "
                    "async/await or provide a safe and separate loop "
                    "to run at the SDK initialization."
                )

        return self._event_loop.run_until_complete(
            _run_awaitable_with_timeout(awaitable, timeout)
        )

    def sync_close(self, timeout: float | None = None) -> None:
        return self.run_sync(self.close(), timeout)

    async def close(self, grace: float | None = None) -> None:
        self._closed = True
        awaits = list[Coroutine[Any, Any, Any]]()
        for chans in self._free_channels.values():
            for chan in chans:
                awaits.append(chan.close(grace))
        for graceful in self._gracefuls:
            awaits.append(graceful.close(grace))
        for task in self._tasks:
            task.cancel()
        rets = await gather(*awaits, *self._tasks, return_exceptions=True)
        for ret in rets:
            if isinstance(ret, BaseException) and not isinstance(ret, CancelledError):
                logger.error(f"Error while graceful shutdown: {ret}", exc_info=ret)

    def get_corresponding_operation_service(
        self,
        service_stub_class: type[ServiceStub],
    ) -> OperationServiceStub:
        addr = self.get_addr_from_stub(service_stub_class)
        chan = self.get_channel_by_addr(addr)
        return OperationServiceStub(chan)  # type: ignore[no-untyped-call]

    def get_corresponding_operation_service_alpha(
        self,
        service_stub_class: type[ServiceStub],
    ) -> OperationServiceStubDeprecated:
        addr = self.get_addr_from_stub(service_stub_class)
        chan = self.get_channel_by_addr(addr)
        return OperationServiceStubDeprecated(chan)  # type: ignore[no-untyped-call]

    def get_addr_from_stub(self, service_stub_class: type[ServiceStub]) -> str:
        service = from_stub_class(service_stub_class)
        return self.get_addr_from_service_name(service)

    def get_addr_from_service_name(self, service_name: str) -> str:
        if len(service_name) > 1 and service_name[0] == ".":
            service_name = service_name[1:]
        return self._resolver.resolve(service_name)

    def get_addr_by_method(self, method_name: str) -> str:
        if method_name not in self._methods:
            service_name = service_from_method_name(method_name)
            self._methods[method_name] = self.get_addr_from_service_name(service_name)
        return self._methods[method_name]

    def get_channel_by_addr(self, addr: str) -> AddressChannel:
        if self._closed:
            raise ChannelClosedError("Channel closed")
        if addr not in self._free_channels:
            self._free_channels[addr] = []
        chans = self._free_channels[addr]
        while len(chans) > 0:
            chan = chans.pop()
            if chan.get_state() != ChannelConnectivity.SHUTDOWN:
                return AddressChannel(chan, addr)
            self.bg_task(chan.close(None))

        return self.create_address_channel(addr)

    def return_channel(self, chan: AddressChannel | None) -> None:
        if chan is None:
            return
        if self._closed:
            raise ChannelClosedError("Channel closed")
        if chan.address not in self._free_channels:
            self._free_channels[chan.address] = []
        if (
            chan.channel.get_state() != ChannelConnectivity.SHUTDOWN
            and len(self._free_channels[chan.address])
            < self._max_free_channels_per_address
        ):
            self._free_channels[chan.address].append(chan.channel)
        else:
            self.discard_channel(chan)

    def discard_channel(self, chan: AddressChannel | None) -> None:
        if chan is None:
            return
        if self._closed:
            raise ChannelClosedError("Channel closed")
        self.bg_task(chan.channel.close(None))

    def get_channel_by_method(self, method_name: str) -> AddressChannel:
        addr = self.get_addr_by_method(method_name)
        return self.get_channel_by_addr(addr)

    def get_address_options(self, addr: str) -> ChannelArgumentType:
        ret = [opt for opt in self._global_options]
        if addr in self._address_options:
            ret.extend(self._address_options[addr])
        ret = set_user_agent_option(self.user_agent, ret)  # type: ignore[assignment]
        return ret

    def get_address_interceptors(self, addr: str) -> Sequence[ClientInterceptor]:
        ret = [opt for opt in self._global_interceptors]
        if addr in self._address_interceptors:
            ret.extend(self._address_interceptors[addr])
        ret.extend(self._global_interceptors_inner)
        return ret

    def create_address_channel(self, addr: str) -> AddressChannel:
        logger.debug(f"creating channel for {addr=}")
        opts = self.get_address_options(addr)
        opts, insecure = pop_option(opts, INSECURE, bool)
        opts, compression = pop_option(opts, COMPRESSION, Compression)
        interceptors = self.get_address_interceptors(addr)
        if insecure:
            return AddressChannel(
                insecure_channel(addr, opts, compression, interceptors),  # type: ignore[unused-ignore,no-any-return]
                addr,
            )
        else:
            return AddressChannel(
                secure_channel(  # type: ignore[unused-ignore,no-any-return]
                    addr,
                    self._tls_credentials,
                    opts,
                    compression,
                    interceptors,
                ),
                addr,
            )

    def unary_unary(  # type: ignore[unused-ignore,override]
        self,
        method_name: str,
        request_serializer: SerializingFunction | None = None,
        response_deserializer: DeserializingFunction | None = None,
    ) -> UnaryUnaryMultiCallable[Req, Res]:  # type: ignore[unused-ignore,override]
        return NebiusUnaryUnaryMultiCallable(
            self,
            method_name,
            request_serializer,
            response_deserializer,
        )

    async def __aenter__(self) -> "Channel":
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close(None)

    def get_state(self, try_to_connect: bool = False) -> ChannelConnectivity:
        if self._closed:
            return ChannelConnectivity.SHUTDOWN
        return ChannelConnectivity.READY

    async def wait_for_state_change(
        self,
        last_observed_state: ChannelConnectivity,
    ) -> None:
        raise NotImplementedError("this method has no meaning for this channel")

    async def channel_ready(self) -> None:
        return

    def unary_stream(  # type: ignore[unused-ignore,override]
        self,
        method: str,
        request_serializer: SerializingFunction | None = None,
        response_deserializer: DeserializingFunction | None = None,
    ) -> UnaryStreamMultiCallable[Req, Res]:  # type: ignore[unused-ignore]
        raise NotImplementedError("Method not implemented")

    def stream_unary(  # type: ignore[unused-ignore,override]
        self,
        method: str,
        request_serializer: SerializingFunction | None = None,
        response_deserializer: DeserializingFunction | None = None,
    ) -> StreamUnaryMultiCallable:
        raise NotImplementedError("Method not implemented")

    def stream_stream(  # type: ignore[unused-ignore,override]
        self,
        method: str,
        request_serializer: SerializingFunction | None = None,
        response_deserializer: DeserializingFunction | None = None,
    ) -> StreamStreamMultiCallable:
        raise NotImplementedError("Method not implemented")



from asyncio import sleep
from datetime import timedelta
from pathlib import Path
from time import time
from typing import IO, Any, AnyStr

from portalocker import Lock as PortalockerLock
from portalocker.constants import LockFlags
from portalocker.exceptions import AlreadyLocked


class Lock:
    def __init__(
        self,
        file_path: str | Path,
        mode: str = "a",
        create_mode: int = 0o644,
        shared: bool = False,
        timeout: timedelta | float | None = None,
        polling_interval: timedelta | float = timedelta(milliseconds=250),
        **fopen_kwargs: Any,
    ):
        self.file_path = Path(file_path)
        self.shared = shared
        self.create_mode = create_mode
        self.timeout = (
            timeout.total_seconds() if isinstance(timeout, timedelta) else timeout
        )
        self.mode: str = mode
        self.fopen_kwargs = fopen_kwargs
        self.polling_interval = (
            polling_interval.total_seconds()
            if isinstance(polling_interval, timedelta)
            else polling_interval
        )
        lock_flags = LockFlags.SHARED if self.shared else LockFlags.EXCLUSIVE
        lock_flags |= LockFlags.NON_BLOCKING
        self.lock = PortalockerLock(
            self.file_path,
            mode=self.mode,  # type: ignore
            timeout=0,
            flags=lock_flags,
            **self.fopen_kwargs,
        )

    async def __aenter__(self) -> IO[AnyStr]:
        start = time()
        while True:
            try:
                try:
                    self.file_path.touch(mode=self.create_mode, exist_ok=False)
                except FileExistsError:
                    pass
                return self.lock.acquire()
            except AlreadyLocked:
                if self.timeout is not None and time() - start > self.timeout:
                    raise TimeoutError(
                        f"Failed to acquire lock on {self.file_path} after "
                        f"{self.timeout} seconds."
                    )
                await sleep(self.polling_interval)
                continue

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: Any,
    ) -> None:
        self.lock.release()



from asyncio import Task
from datetime import datetime, timedelta, timezone
from logging import getLogger
from ssl import SSLContext
from typing import Any, TextIO, TypeVar

from nebius.aio.token.token import Bearer as ParentBearer
from nebius.aio.token.token import Receiver as ParentReceiver
from nebius.aio.token.token import Token

log = getLogger(__name__)


class Receiver(ParentReceiver):
    def __init__(
        self,
        client_id: str,
        federation_endpoint: str,
        federation_id: str,
        writer: TextIO | None = None,
        no_browser_open: bool = False,
        ssl_ctx: SSLContext | None = None,
    ) -> None:
        self._client_id = client_id
        self._federation_endpoint = federation_endpoint
        self._federation_id = federation_id
        self._writer = writer
        self._no_browser_open = no_browser_open
        self._ssl_ctx = ssl_ctx

    async def _fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        from .auth import authorize

        now = datetime.now(timezone.utc)
        tok = await authorize(
            client_id=self._client_id,
            federation_endpoint=self._federation_endpoint,
            federation_id=self._federation_id,
            writer=self._writer,
            no_browser_open=self._no_browser_open,
            timeout=timeout,
            ssl_ctx=self._ssl_ctx,
        )
        return Token(
            token=tok.access_token,
            expiration=(
                now + timedelta(seconds=tok.expires_in)
                if tok.expires_in is not None
                else None
            ),
        )

    def can_retry(
        self,
        err: Exception,
        options: dict[str, str] | None = None,
    ) -> bool:
        return True


T = TypeVar("T")


class Bearer(ParentBearer):
    def __init__(
        self,
        profile_name: str,
        client_id: str,
        federation_endpoint: str,
        federation_id: str,
        writer: TextIO | None = None,
        no_browser_open: bool = False,
        ssl_ctx: SSLContext | None = None,
    ) -> None:
        self._profile_name = profile_name
        self._client_id = client_id
        self._federation_endpoint = federation_endpoint
        self._federation_id = federation_id
        self._writer = writer
        self._no_browser_open = no_browser_open
        self._ssl_ctx = ssl_ctx

        self._tasks = set[Task[Any]]()

    @property
    def name(self) -> str:
        return (
            f"federation/{self._federation_endpoint}/{self._federation_id}/"
            f"{self._profile_name}"
        )

    def receiver(self) -> ParentReceiver:
        return Receiver(
            client_id=self._client_id,
            federation_endpoint=self._federation_endpoint,
            federation_id=self._federation_id,
            writer=self._writer,
            no_browser_open=self._no_browser_open,
            ssl_ctx=self._ssl_ctx,
        )



from nebius.api.nebius import annotations_pb2 as _annotations_pb2
from nebius.api.nebius.compute.v1 import maintenance_event_pb2 as _maintenance_event_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetMaintenanceEventByInstanceRequest(_message.Message):
    __slots__ = ["instance_id"]
    INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    instance_id: str
    def __init__(self, instance_id: _Optional[str] = ...) -> None: ...

class ListMaintenanceEventsRequest(_message.Message):
    __slots__ = ["parent_id", "page_size", "page_token"]
    PARENT_ID_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    parent_id: str
    page_size: int
    page_token: str
    def __init__(self, parent_id: _Optional[str] = ..., page_size: _Optional[int] = ..., page_token: _Optional[str] = ...) -> None: ...

class ListMaintenanceEventsResponse(_message.Message):
    __slots__ = ["items", "next_page_token"]
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    items: _containers.RepeatedCompositeFieldContainer[_maintenance_event_pb2.MaintenanceEvent]
    next_page_token: str
    def __init__(self, items: _Optional[_Iterable[_Union[_maintenance_event_pb2.MaintenanceEvent, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...



from asyncio import CancelledError, Future, ensure_future, wait_for
from collections.abc import Awaitable, Callable, Generator, Iterable
from logging import getLogger
from sys import exc_info
from time import time
from typing import Any, Generic, TypeVar

from google.protobuf.message import Message as PMessage
from grpc import CallCredentials, Compression, StatusCode
from grpc.aio import AioRpcError
from grpc.aio import Metadata as GrpcMetadata
from grpc.aio._call import UnaryUnaryCall  # type: ignore[unused-ignore]
from grpc_status import rpc_status

from nebius.aio.abc import ClientChannelInterface as Channel
from nebius.aio.authorization.options import OPTION_TYPE, Types
from nebius.aio.base import AddressChannel
from nebius.aio.idempotency import ensure_key_in_metadata
from nebius.base.error import SDKError
from nebius.base.metadata import Metadata
from nebius.base.protos.unset import Unset, UnsetType

from .request_status import RequestStatus, UnfinishedRequestStatus

Req = TypeVar("Req")
Res = TypeVar("Res")
Err = TypeVar("Err")
T = TypeVar("T")

log = getLogger(__name__)


class RequestError(SDKError):
    pass


class RequestIsSentError(RequestError):
    def __init__(self) -> None:
        super().__init__("Request is already sent")


class RequestIsCancelledError(RequestError):
    def __init__(self) -> None:
        super().__init__("Request is cancelled")


class RequestSentNoCallError(RequestError):
    def __init__(self) -> None:
        super().__init__("Request marked as sent without call.")


DEFAULT_TIMEOUT = 60.0  # second
DEFAULT_PER_RETRY_TIMEOUT = DEFAULT_TIMEOUT / 3
DEFAULT_AUTH_TIMEOUT = 15 * 60.0  # 15 minutes


class Request(Generic[Req, Res]):
    def __init__(
        self,
        channel: Channel,
        service: str,
        method: str,
        request: Req,
        result_pb2_class: type[PMessage],
        metadata: Metadata | Iterable[tuple[str, str]] | None = None,
        timeout: float | None | UnsetType = Unset,
        auth_timeout: float | None | UnsetType = Unset,
        auth_options: dict[str, str] | None = None,
        credentials: CallCredentials | None = None,
        compression: Compression | None = None,
        result_wrapper: Callable[[str, Channel, Any], Res] | None = None,
        grpc_channel_override: AddressChannel | None = None,
        error_wrapper: Callable[[RequestStatus], RequestError] | None = None,
        retries: int | None = 3,
        per_retry_timeout: float | None | UnsetType = Unset,
    ) -> None:
        self._channel = channel
        self._input = request
        self._service = service
        self._method = method
        self._auth_options = auth_options if auth_options is not None else {}
        self._result_pb2_class = result_pb2_class
        self._input_metadata = Metadata(metadata)
        self._result_wrapper = result_wrapper
        self._grpc_channel = grpc_channel_override
        self._timeout: float | None = (
            timeout if not isinstance(timeout, UnsetType) else DEFAULT_TIMEOUT
        )
        self._per_retry_timeout: float | None = (
            per_retry_timeout
            if not isinstance(per_retry_timeout, UnsetType)
            else DEFAULT_PER_RETRY_TIMEOUT
        )
        self._auth_timeout: float | None = (
            auth_timeout
            if not isinstance(auth_timeout, UnsetType)
            else DEFAULT_AUTH_TIMEOUT
        )
        self._credentials = credentials
        self._compression = compression
        self._call: UnaryUnaryCall | None = None
        self._retries = retries
        self._cancelled: bool = False
        from .service_error import RequestError as RSError
        from .service_error import RequestStatusExtended

        ensure_key_in_metadata(self._input_metadata)

        self._error_wrapper = error_wrapper if error_wrapper is not None else RSError
        self._status: RequestStatusExtended | None = None
        self._initial_metadata: Metadata | None = None
        self._trailing_metadata: Metadata | None = None
        self._trace_id: str | None = None
        self._request_id: str | None = None

        self._awaited = False
        self._future: Future[Res] | None = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self._service}.{self._method}, "
            f"{self.current_status()})"
        )

    def done(self) -> bool:
        if self._call is None:
            return False
        return self._call.done()

    def cancelled(self) -> bool:
        if self._call is not None:
            return self._call.cancelled()
        return self._cancelled

    def cancel(self) -> bool:
        if self._call is not None:
            return self._call.cancel()
        else:
            self._cancelled = True
            return self._cancelled

    def input_metadata(self) -> Metadata:
        return self._input_metadata

    @property
    def timeout(self) -> float | None:
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: float | None) -> None:
        if self._call is not None:
            raise RequestIsSentError()
        self._timeout = timeout

    @property
    def credentials(self) -> CallCredentials | None:
        return self._credentials

    @credentials.setter
    def credentials(self, credentials: CallCredentials | None) -> None:
        if self._call is not None:
            raise RequestIsSentError()
        self._credentials = credentials

    @property
    def wait_for_ready(self) -> bool | None:
        return self._wait_for_ready

    @wait_for_ready.setter
    def wait_for_ready(self, wait_for_ready: bool | None) -> None:
        if self._call is not None:
            raise RequestIsSentError()
        self._wait_for_ready = wait_for_ready

    @property
    def compression(self) -> Compression | None:
        return self._compression

    @compression.setter
    def compression(self, compression: Compression | None) -> None:
        if self._call is not None:
            raise RequestIsSentError()
        self._compression = compression

    def _send(self, timeout: float | None) -> None:
        from nebius.base.protos.pb_classes import Message

        self._initial_metadata = None
        self._trailing_metadata = None
        self._status = None
        req = self._input
        if isinstance(req, Message):
            req = req.__pb2_message__  # type: ignore[assignment]
        if isinstance(req, PMessage):
            serializer = req.__class__.SerializeToString
        else:
            raise RequestError(f"Unsupported request type {type(req)}")
        if self._cancelled:
            raise RequestIsCancelledError()
        channel_parent_id = self._channel.parent_id()
        if channel_parent_id is not None:
            if self._method == "List" or self._method == "GetByName":
                if hasattr(req, "parent_id") and req.parent_id == "":  # type: ignore[unused-ignore]
                    req.parent_id = channel_parent_id  # type: ignore[unused-ignore]
            elif self._method != "Update":
                if hasattr(req, "metadata") and hasattr(req.metadata, "parent_id"):  # type: ignore[unused-ignore]
                    if req.metadata.parent_id == "":  # type: ignore[unused-ignore]
                        req.metadata.parent_id = channel_parent_id  # type: ignore[unused-ignore]
        self._sent = True
        if self._grpc_channel is None:
            self._grpc_channel = self._channel.get_channel_by_method(
                self._service + "." + self._method
            )
        s_name = self._service
        if s_name[0] == ".":
            s_name = s_name[1:]
        self._call = self._grpc_channel.channel.unary_unary(  # type: ignore
            "/" + s_name + "/" + self._method,
            serializer,
            self._result_pb2_class.FromString,
        )(
            req,
            timeout=timeout,
            metadata=GrpcMetadata(*self._input_metadata),
            credentials=self._credentials,
            wait_for_ready=True,
            compression=self._compression,
        )

    def run_sync_with_timeout(self, func: Awaitable[T]) -> T:
        try:
            # Overall timeout should include authorization + request execution
            timeout = self._auth_timeout
            if timeout is not None:
                timeout += 0.2  # 200 ms for an internal graceful shutdown
            return self._channel.run_sync(func, timeout=timeout)
        except TimeoutError as e:
            from .service_error import RequestError, RequestStatusExtended

            self._status = RequestStatusExtended(
                code=StatusCode.DEADLINE_EXCEEDED,
                message="Deadline Exceeded",
                details=[],
                service_errors=[],
                request_id=self._request_id if self._request_id is not None else "",
                trace_id=self._trace_id if self._trace_id is not None else "",
            )
            raise RequestError(self._status) from e

    def wait(self) -> Res:
        return self.run_sync_with_timeout(self)

    def initial_metadata_sync(self) -> Metadata:
        if self._initial_metadata is not None:
            return self._initial_metadata
        return self.run_sync_with_timeout(self.initial_metadata())

    def trailing_metadata_sync(self) -> Metadata:
        if self._trailing_metadata is not None:
            return self._trailing_metadata
        return self.run_sync_with_timeout(self.trailing_metadata())

    def current_status(self) -> RequestStatus | UnfinishedRequestStatus:
        if self._status is not None:
            return self._status
        if self._call is None:
            return UnfinishedRequestStatus.INITIALIZED
        return UnfinishedRequestStatus.SENT

    async def _get_request_id(self) -> tuple[str, str]:
        if self._request_id is not None and self._trace_id is not None:
            return (self._request_id, self._trace_id)
        await self.initial_metadata()
        return (self._request_id, self._trace_id)  # type: ignore[return-value] # should be set after receiving md

    async def request_id(self) -> str:
        ret = await self._get_request_id()
        return ret[0]

    async def trace_id(self) -> str:
        ret = await self._get_request_id()
        return ret[1]

    def request_id_sync(self) -> str:
        if self._request_id is not None:
            return self._request_id
        return self.run_sync_with_timeout(self.request_id())

    def trace_id_sync(self) -> str:
        if self._trace_id is not None:
            return self._trace_id
        return self.run_sync_with_timeout(self.trace_id())

    async def initial_metadata(self) -> Metadata:
        try:
            await self._await_result()
        except Exception as e:  # noqa: S110
            if self._initial_metadata is not None:
                return self._initial_metadata
            raise e
        if self._initial_metadata is not None:
            return self._initial_metadata
        raise RequestError("no initial metadata after call finished")

    async def trailing_metadata(self) -> Metadata:
        try:
            await self._await_result()
        except Exception as e:  # noqa: S110
            if self._trailing_metadata is not None:
                return self._trailing_metadata
            raise e
        if self._trailing_metadata is not None:
            return self._trailing_metadata
        raise RequestError("no trailing metadata after call finished")

    def _parse_request_id(self) -> None:
        if self._initial_metadata is None:
            raise RequestError("no initial metadata")
        self._request_id = self._initial_metadata.get_one("x-request-id", "")
        self._trace_id = self._initial_metadata.get_one("x-trace-id", "")

    async def status(self) -> RequestStatus:
        try:
            await self._await_result()
        except Exception as e:  # noqa: S110
            if self._status is not None:
                return self._status
            raise e
        if self._status is not None:
            return self._status
        raise RequestError("no status after call finished")

    def _raise_request_error(self, err: AioRpcError) -> None:
        self._initial_metadata = Metadata(err.initial_metadata())
        self._trailing_metadata = Metadata(err.trailing_metadata())  # type: ignore
        self._parse_request_id()
        status = rpc_status.from_call(err)  # type: ignore
        from .service_error import RequestError, RequestStatusExtended

        debug_info = err.debug_error_string()
        if debug_info:
            log.debug(f"RPC Debug info: {debug_info}")

        if status is None:
            self._status = RequestStatusExtended(
                code=err.code(),
                message=err.details(),
                details=[],
                service_errors=[],
                request_id=self._request_id,  # type: ignore[arg-type] # should be strings by now
                trace_id=self._trace_id,  # type: ignore[arg-type] # should be strings by now
            )
            raise RequestError(self._status) from None

        self._status = RequestStatusExtended.from_rpc_status(  # type: ignore[unused-ignore]
            status,
            trace_id=self._trace_id,  # type: ignore[arg-type] # should be strings by now
            request_id=self._request_id,  # type: ignore[arg-type] # should be known by now
        )
        raise RequestError(self._status) from None

    def _convert_request_error(self, err: AioRpcError) -> None:
        from .service_error import RequestError

        try:
            self._raise_request_error(err)
        except RequestError:
            pass

    async def _retry_loop(self, outer_deadline: float | None = None) -> Res:
        from .service_error import RequestError, is_retriable_error

        self._start_time = time()
        # Compute this loop's absolute deadline, capped by an outer deadline if provided
        own_deadline = (
            None if self._timeout is None else self._start_time + self._timeout
        )
        if outer_deadline is None:
            deadline = own_deadline
        elif own_deadline is None:
            deadline = outer_deadline
        else:
            deadline = min(own_deadline, outer_deadline)
        attempt = 0
        while not self._cancelled:
            attempt += 1
            timeout = None if deadline is None else deadline - time()

            if self._per_retry_timeout is not None and (
                timeout is None or timeout > self._per_retry_timeout
            ):
                # Clip per-retry timeout by remaining overall deadline if present
                per_attempt = self._per_retry_timeout
                if deadline is not None:
                    remaining = deadline - time()
                    if remaining <= 0:
                        per_attempt = 0
                    else:
                        per_attempt = min(per_attempt, remaining)
                timeout = per_attempt

            # somehow, this time python doesn't want to catch the raised error again
            # thus, it will be two nested try/except blocks
            try:
                try:
                    self._send(timeout)
                    if self._call is None:
                        raise RequestSentNoCallError()
                    ret = await self._call  # type: ignore[unused-ignore]
                    code = await self._call.code()
                    msg = await self._call.details()
                    mdi = await self._call.initial_metadata()
                    mdt = await self._call.trailing_metadata()
                    e = AioRpcError(code, mdi, mdt, msg, None)  # type: ignore
                    self._convert_request_error(e)
                    if self._result_wrapper is not None:
                        return self._result_wrapper(
                            self._service + "." + self._method,
                            self._channel,
                            ret,
                        )
                    self._channel.return_channel(self._grpc_channel)
                    self._grpc_channel = None
                    return ret  # type: ignore
                except CancelledError as e:
                    self._channel.discard_channel(self._grpc_channel)
                    self._grpc_channel = None
                    raise e
                except AioRpcError as e:
                    self._raise_request_error(e)
            except Exception as e:
                if (
                    (deadline is None or deadline > time())
                    and is_retriable_error(e, deadline_retriable=True)
                    and (self._retries is None or self._retries > attempt)
                ):
                    log.error(
                        f"request attempt {attempt} for {self} failed with {e} "
                        + "but will be retried",
                        exc_info=exc_info(),
                    )
                    continue
                if deadline is not None and deadline <= time():
                    if isinstance(e, RequestError) and e.status.request_id == "":
                        self._channel.discard_channel(self._grpc_channel)
                        self._grpc_channel = None
                self._channel.return_channel(self._grpc_channel)
                raise e
        raise RequestIsCancelledError()

    async def _request_with_authorization_loop(self) -> Res:
        """Wrap request retry loop with an authorization loop.

        The authorization loop will attempt to authenticate and then execute the
        request retry loop. If the result is UNAUTHENTICATED and the authenticator
        allows retry, it will re-authenticate and try again while respecting the
        overall auth timeout.
        """
        # If no provider or authorization explicitly disabled, just run the request
        provider = self._channel.get_authorization_provider()

        auth_type = self._auth_options.get(OPTION_TYPE, None)
        if provider is None or auth_type == Types.DISABLE:
            return await self._retry_loop()

        start = time()
        deadline = None if self._auth_timeout is None else start + self._auth_timeout
        attempt = 0
        auth = provider.authenticator()

        while True:
            attempt += 1
            timeout = None if deadline is None else (deadline - time())
            if timeout is not None and timeout <= 0:
                raise TimeoutError("authorization timed out")
            # Perform authentication: use a gRPC Metadata, then copy back auth header
            # to the internal Metadata used when sending the request
            auth_md = Metadata(self._input_metadata)
            try:
                await wait_for(
                    auth.authenticate(auth_md, timeout, self._auth_options), timeout
                )
            except Exception as e:  # noqa: BLE001
                # If authentication itself failed (e.g., token refresh timeout),
                # retry if authenticator allows and we are within deadline.
                if deadline is not None and deadline <= time():
                    raise
                if auth.can_retry(e, self._auth_options):
                    continue
                raise
            self._input_metadata = auth_md

            # Run the request retry loop; map UNAUTHENTICATED to re-auth attempts
            try:
                return await self._retry_loop(outer_deadline=deadline)
            except Exception as e:  # noqa: BLE001
                # Only retry auth on UNAUTHENTICATED codes
                from .service_error import RequestError as ServiceRequestError

                if not isinstance(e, ServiceRequestError):
                    raise
                try:
                    code = e.status.code
                except Exception:
                    # If code is not available, don't treat it as auth failure
                    raise
                if not (
                    code == StatusCode.UNAUTHENTICATED
                    or getattr(code, "name", None) == "UNAUTHENTICATED"
                ):
                    raise
                if deadline is not None and deadline <= time():
                    raise
                if not auth.can_retry(e, self._auth_options):
                    raise
                # loop continues to re-authenticate and retry

    async def _await_result(self) -> Res:
        if self._future is None:
            self._future = ensure_future(self._request_with_authorization_loop())
        return await self._future

    def __await__(self) -> Generator[Any, None, Res]:
        if self._awaited:
            raise RuntimeError("cannot await the finished coroutine")
        self._awaited = True

        res = yield from self._await_result().__await__()
        return res



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Condition(_message.Message):
    __slots__ = ["type", "status", "last_transition_at", "reason", "severity", "description", "last_transition_error"]
    class Severity(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        NONE: _ClassVar[Condition.Severity]
        INFO: _ClassVar[Condition.Severity]
        ERROR: _ClassVar[Condition.Severity]
    NONE: Condition.Severity
    INFO: Condition.Severity
    ERROR: Condition.Severity
    class Status(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        UNKNOWN: _ClassVar[Condition.Status]
        TRUE: _ClassVar[Condition.Status]
        FALSE: _ClassVar[Condition.Status]
    UNKNOWN: Condition.Status
    TRUE: Condition.Status
    FALSE: Condition.Status
    class TransitionError(_message.Message):
        __slots__ = ["reason", "description"]
        REASON_FIELD_NUMBER: _ClassVar[int]
        DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
        reason: str
        description: str
        def __init__(self, reason: _Optional[str] = ..., description: _Optional[str] = ...) -> None: ...
    TYPE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    LAST_TRANSITION_AT_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    SEVERITY_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    LAST_TRANSITION_ERROR_FIELD_NUMBER: _ClassVar[int]
    type: str
    status: Condition.Status
    last_transition_at: _timestamp_pb2.Timestamp
    reason: str
    severity: Condition.Severity
    description: str
    last_transition_error: Condition.TransitionError
    def __init__(self, type: _Optional[str] = ..., status: _Optional[_Union[Condition.Status, str]] = ..., last_transition_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., reason: _Optional[str] = ..., severity: _Optional[_Union[Condition.Severity, str]] = ..., description: _Optional[str] = ..., last_transition_error: _Optional[_Union[Condition.TransitionError, _Mapping]] = ...) -> None: ...



from datetime import timedelta
from ssl import SSLContext
from typing import TextIO

from nebius.aio.token.token import Bearer as ParentBearer
from nebius.aio.token.token import Receiver

from .federation_bearer import Bearer as FederationAuthBearer
from .file_cache.async_renewable_bearer import AsynchronousRenewableFileCacheBearer


class FederationBearer(ParentBearer):
    def __init__(
        self,
        profile_name: str,
        client_id: str,
        federation_endpoint: str,
        federation_id: str,
        writer: TextIO | None = None,
        no_browser_open: bool = False,
        timeout: timedelta = timedelta(minutes=5),
        max_retries: int = 2,
        initial_safety_margin: timedelta | float | None = timedelta(hours=2),
        retry_safety_margin: timedelta = timedelta(hours=2),
        lifetime_safe_fraction: float = 0.9,
        initial_retry_timeout: timedelta = timedelta(seconds=1),
        max_retry_timeout: timedelta = timedelta(minutes=1),
        retry_timeout_exponent: float = 1.5,
        file_cache_throttle: timedelta | float = timedelta(minutes=5),
        ssl_ctx: SSLContext | None = None,
    ) -> None:
        self._source = AsynchronousRenewableFileCacheBearer(
            FederationAuthBearer(
                profile_name=profile_name,
                client_id=client_id,
                federation_endpoint=federation_endpoint,
                federation_id=federation_id,
                writer=writer,
                no_browser_open=no_browser_open,
                ssl_ctx=ssl_ctx,
            ),
            max_retries=max_retries,
            initial_safety_margin=initial_safety_margin,
            retry_safety_margin=retry_safety_margin,
            lifetime_safe_fraction=lifetime_safe_fraction,
            initial_retry_timeout=initial_retry_timeout,
            max_retry_timeout=max_retry_timeout,
            retry_timeout_exponent=retry_timeout_exponent,
            refresh_request_timeout=timeout,
            file_cache_throttle=file_cache_throttle,
        )

    @property
    def wrapped(self) -> "ParentBearer|None":
        return self._source

    def receiver(self) -> "Receiver":
        return self._source.receiver()
